"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import time
import logging
import uuid
import os

from utils import retry_limited
import pytest
from inmanta.agent.agent import Agent
from inmanta import data, const, config
from inmanta.server import config as opt, SLICE_AGENT_MANAGER, SLICE_SESSION_MANAGER, server
from datetime import datetime
from uuid import UUID
from inmanta.util import hash_file
from inmanta.export import upload_code, unknown_parameters
import asyncio


LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio(timeout=60)
@pytest.mark.slowtest
async def test_autostart(server, client, environment):
    """
        Test auto start of agent
    """
    env = await data.Environment.get_by_id(uuid.UUID(environment))
    await env.set(data.AUTOSTART_AGENT_MAP, {"iaas_agent": "", "iaas_agentx": ""})

    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)
    sessionendpoint = server.get_slice(SLICE_SESSION_MANAGER)

    await agentmanager.ensure_agent_registered(env, "iaas_agent")
    await agentmanager.ensure_agent_registered(env, "iaas_agentx")

    res = await agentmanager._ensure_agents(env, ["iaas_agent"])
    assert res
    await retry_limited(lambda: len(sessionendpoint._sessions) == 1, 20)
    assert len(sessionendpoint._sessions) == 1
    res = await agentmanager._ensure_agents(env, ["iaas_agent"])
    assert not res
    assert len(sessionendpoint._sessions) == 1

    LOGGER.warning("Killing agent")
    agentmanager._agent_procs[env.id].proc.terminate()
    await agentmanager._agent_procs[env.id].wait_for_exit(raise_error=False)
    await retry_limited(lambda: len(sessionendpoint._sessions) == 0, 20)
    res = await agentmanager._ensure_agents(env, ["iaas_agent"])
    assert res
    await retry_limited(lambda: len(sessionendpoint._sessions) == 1, 3)
    assert len(sessionendpoint._sessions) == 1

    # second agent for same env
    res = await agentmanager._ensure_agents(env, ["iaas_agentx"])
    assert res
    await retry_limited(lambda: len(sessionendpoint._sessions) == 1, 20)
    assert len(sessionendpoint._sessions) == 1

    # Test stopping all agents
    await agentmanager.stop_agents(env)
    assert len(sessionendpoint._sessions) == 0
    assert len(agentmanager._agent_procs) == 0


@pytest.mark.asyncio(timeout=60)
@pytest.mark.slowtest
async def test_autostart_dual_env(client, server):
    """
        Test auto start of agent
    """
    agentmanager = server.get_slice("server").agentmanager
    sessionendpoint = server.get_slice("session")

    result = await client.create_project("env-test")
    assert result.code == 200
    project_id = result.result["project"]["id"]

    result = await client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    result = await client.create_environment(project_id=project_id, name="devx")
    env_id2 = result.result["environment"]["id"]

    env = await data.Environment.get_by_id(uuid.UUID(env_id))
    await env.set(data.AUTOSTART_AGENT_MAP, {"iaas_agent": ""})

    env2 = await data.Environment.get_by_id(uuid.UUID(env_id2))
    await env2.set(data.AUTOSTART_AGENT_MAP, {"iaas_agent": ""})

    await agentmanager.ensure_agent_registered(env, "iaas_agent")
    await agentmanager.ensure_agent_registered(env2, "iaas_agent")

    res = await agentmanager._ensure_agents(env, ["iaas_agent"])
    assert res
    await retry_limited(lambda: len(sessionendpoint._sessions) == 1, 20)
    assert len(sessionendpoint._sessions) == 1

    res = await agentmanager._ensure_agents(env2, ["iaas_agent"])
    assert res
    await retry_limited(lambda: len(sessionendpoint._sessions) == 2, 20)
    assert len(sessionendpoint._sessions) == 2


@pytest.mark.asyncio(timeout=60)
@pytest.mark.slowtest
async def test_autostart_batched(client, server, environment):
    """
        Test auto start of agent
    """
    env = await data.Environment.get_by_id(uuid.UUID(environment))
    await env.set(data.AUTOSTART_AGENT_MAP, {"iaas_agent": "", "iaas_agentx": ""})

    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)
    sessionendpoint = server.get_slice(SLICE_SESSION_MANAGER)

    await agentmanager.ensure_agent_registered(env, "iaas_agent")
    await agentmanager.ensure_agent_registered(env, "iaas_agentx")

    res = await agentmanager._ensure_agents(env, ["iaas_agent", "iaas_agentx"])
    assert res
    await retry_limited(lambda: len(sessionendpoint._sessions) == 1, 20)
    assert len(sessionendpoint._sessions) == 1
    res = await agentmanager._ensure_agents(env, ["iaas_agent"])
    assert not res
    assert len(sessionendpoint._sessions) == 1

    res = await agentmanager._ensure_agents(env, ["iaas_agent", "iaas_agentx"])
    assert not res
    assert len(sessionendpoint._sessions) == 1

    LOGGER.warning("Killing agent")
    agentmanager._agent_procs[env.id].proc.terminate()
    await agentmanager._agent_procs[env.id].wait_for_exit(raise_error=False)
    await retry_limited(lambda: len(sessionendpoint._sessions) == 0, 20)
    res = await agentmanager._ensure_agents(env, ["iaas_agent", "iaas_agentx"])
    assert res
    await retry_limited(lambda: len(sessionendpoint._sessions) == 1, 3)
    assert len(sessionendpoint._sessions) == 1


@pytest.mark.asyncio(timeout=10)
async def test_version_removal(client, server):
    """
        Test auto removal of older deploy model versions
    """
    result = await client.create_project("env-test")
    assert result.code == 200
    project_id = result.result["project"]["id"]

    result = await client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    version = int(time.time())

    for _i in range(20):
        version += 1

        await server.get_slice("server")._purge_versions()
        res = await client.put_version(tid=env_id, version=version, resources=[], unknowns=[], version_info={})
        assert res.code == 200
        result = await client.get_project(id=project_id)

        versions = await client.list_versions(tid=env_id)
        assert versions.result["count"] <= opt.server_version_to_keep.get() + 1


@pytest.mark.asyncio(timeout=30)
@pytest.mark.slowtest
async def test_get_resource_for_agent(motor, server_multi, client_multi, environment):
    """
        Test the server to manage the updates on a model during agent deploy
    """
    agent = Agent("localhost", {"nvblah": "localhost"}, environment=environment, code_loader=False)
    await agent.start()
    aclient = agent._client

    version = 1

    resources = [{'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1.dev.inmanta.com,path=/etc/sysconfig/network],v=%d' % version,
                  'owner': 'root',
                  'path': '/etc/sysconfig/network',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'version': version},
                 {'group': 'root',
                  'hash': 'b4350bef50c3ec3ee532d4a3f9d6daedec3d2aba',
                  'id': 'std::File[vm2.dev.inmanta.com,path=/etc/motd],v=%d' % version,
                  'owner': 'root',
                  'path': '/etc/motd',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'version': version},
                 {'group': 'root',
                  'hash': '3bfcdad9ab7f9d916a954f1a96b28d31d95593e4',
                  'id': 'std::File[vm1.dev.inmanta.com,path=/etc/hostname],v=%d' % version,
                  'owner': 'root',
                  'path': '/etc/hostname',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'version': version},
                 {'id': 'std::Service[vm1.dev.inmanta.com,name=network],v=%d' % version,
                  'name': 'network',
                  'onboot': True,
                  'requires': ['std::File[vm1.dev.inmanta.com,path=/etc/sysconfig/network],v=%d' % version],
                  'state': 'running',
                  'version': version}]

    res = await client_multi.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert res.code == 200

    result = await client_multi.list_versions(environment)
    assert result.code == 200
    assert result.result["count"] == 1

    result = await client_multi.release_version(environment, version, False)
    assert result.code == 200

    result = await client_multi.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["version"] == version
    assert result.result["model"]["total"] == len(resources)
    assert result.result["model"]["released"]
    assert result.result["model"]["result"] == "deploying"

    result = await aclient.get_resources_for_agent(environment, "vm1.dev.inmanta.com")
    assert result.code == 200
    assert len(result.result["resources"]) == 3

    action_id = uuid.uuid4()
    now = datetime.now()
    result = await aclient.resource_action_update(environment,
                                                  ["std::File[vm1.dev.inmanta.com,path=/etc/sysconfig/network],v=%d" % version],
                                                  action_id, "deploy", now, now, "deployed", [], {})

    assert result.code == 200

    result = await client_multi.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["done"] == 1

    action_id = uuid.uuid4()
    now = datetime.now()
    result = await aclient.resource_action_update(environment,
                                                  ["std::File[vm1.dev.inmanta.com,path=/etc/hostname],v=%d" % version],
                                                  action_id, "deploy", now, now, "deployed", [], {})
    assert result.code == 200

    result = await client_multi.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["done"] == 2
    await agent.stop()


@pytest.mark.asyncio(timeout=10)
async def test_get_environment(client, server, environment):
    version = int(time.time())

    for i in range(10):
        version += 1

        resources = []
        for j in range(i):
            resources.append({
                'group': 'root',
                'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                'id': 'std::File[vm1.dev.inmanta.com,path=/tmp/file%d],v=%d' % (j, version),
                'owner': 'root',
                'path': '/tmp/file%d' % j,
                'permissions': 644,
                'purged': False,
                'reload': False,
                'requires': [],
                'version': version})

        res = await client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
        assert res.code == 200

    result = await client.get_environment(environment, versions=5, resources=1)
    assert(result.code == 200)
    assert(len(result.result["environment"]["versions"]) == 5)
    assert(len(result.result["environment"]["resources"]) == 9)


@pytest.mark.asyncio
async def test_resource_update(client, server, environment):
    """
        Test updating resources and logging
    """
    agent = Agent("localhost", {"blah": "localhost"}, environment=environment, code_loader=False)
    await agent.start()
    aclient = agent._client

    version = int(time.time())

    resources = []
    for j in range(10):
        resources.append({
            'group': 'root',
            'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
            'id': 'std::File[vm1,path=/tmp/file%d],v=%d' % (j, version),
            'owner': 'root',
            'path': '/tmp/file%d' % j,
            'permissions': 644,
            'purged': False,
            'reload': False,
            'requires': [],
            'version': version})

    res = await client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert(res.code == 200)

    result = await client.release_version(environment, version, False)
    assert result.code == 200

    resource_ids = [x["id"] for x in resources]

    # Start the deploy
    action_id = uuid.uuid4()
    now = datetime.now()
    result = await aclient.resource_action_update(environment, resource_ids, action_id, "deploy", now,
                                                  status=const.ResourceState.deployed)
    assert(result.code == 200)

    # Get the status from a resource
    result = await client.get_resource(tid=environment, id=resource_ids[0], logs=True)
    assert(result.code == 200)
    logs = {x["action"]: x for x in result.result["logs"]}

    assert("deploy" in logs)
    assert("finished" not in logs["deploy"])
    assert("messages" not in logs["deploy"])
    assert("changes" not in logs["deploy"])

    # Send some logs
    result = await aclient.resource_action_update(environment, resource_ids, action_id, "deploy",
                                                  status=const.ResourceState.deployed,
                                                  messages=[data.LogLine.log(const.LogLevel.INFO,
                                                                             "Test log %(a)s %(b)s", a="a", b="b")])
    assert(result.code == 200)

    # Get the status from a resource
    result = await client.get_resource(tid=environment, id=resource_ids[0], logs=True)
    assert(result.code == 200)
    logs = {x["action"]: x for x in result.result["logs"]}

    assert("deploy" in logs)
    assert("messages" in logs["deploy"])
    assert(len(logs["deploy"]["messages"]) == 1)
    assert(logs["deploy"]["messages"][0]["msg"] == "Test log a b")
    assert("finished" not in logs["deploy"])
    assert("changes" not in logs["deploy"])

    # Finish the deploy
    now = datetime.now()
    changes = {x: {"owner": {"old": "root", "current": "inmanta"}} for x in resource_ids}
    result = await aclient.resource_action_update(environment, resource_ids, action_id, "deploy", finished=now, changes=changes)
    assert(result.code == 500)

    result = await aclient.resource_action_update(environment, resource_ids, action_id, "deploy", status="deployed",
                                                  finished=now, changes=changes)
    assert(result.code == 200)

    result = await client.get_version(environment, version)
    assert(result.code == 200)
    assert result.result["model"]["done"] == 10
    await agent.stop()


@pytest.mark.asyncio
async def test_environment_settings(client, server, environment):
    """
        Test environment settings
    """
    result = await client.list_settings(tid=environment)
    assert result.code == 200
    assert "settings" in result.result
    assert "metadata" in result.result
    assert "auto_deploy" in result.result["metadata"]
    assert len(result.result["settings"]) == 0

    result = await client.set_setting(tid=environment, id="auto_deploy", value="test")
    assert result.code == 500

    result = await client.set_setting(tid=environment, id="auto_deploy", value=False)
    assert result.code == 200

    result = await client.list_settings(tid=environment)
    assert result.code == 200
    assert len(result.result["settings"]) == 1

    result = await client.get_setting(tid=environment, id="auto_deploy")
    assert result.code == 200
    assert not result.result["value"]

    result = await client.get_setting(tid=environment, id="test2")
    assert result.code == 404

    result = await client.set_setting(tid=environment, id="auto_deploy", value=True)
    assert result.code == 200

    result = await client.get_setting(tid=environment, id="auto_deploy")
    assert result.code == 200
    assert result.result["value"]

    result = await client.delete_setting(tid=environment, id="test2")
    assert result.code == 404

    result = await client.delete_setting(tid=environment, id="auto_deploy")
    assert result.code == 200

    result = await client.list_settings(tid=environment)
    assert result.code == 200
    assert "settings" in result.result
    assert len(result.result["settings"]) == 1

    result = await client.set_setting(tid=environment, id=data.AUTOSTART_AGENT_DEPLOY_SPLAY_TIME, value=20)
    assert result.code == 200

    result = await client.set_setting(tid=environment, id=data.AUTOSTART_AGENT_DEPLOY_SPLAY_TIME, value="30")
    assert result.code == 200

    result = await client.get_setting(tid=environment, id=data.AUTOSTART_AGENT_DEPLOY_SPLAY_TIME)
    assert result.code == 200
    assert result.result["value"] == 30

    result = await client.delete_setting(tid=environment, id=data.AUTOSTART_AGENT_DEPLOY_SPLAY_TIME)
    assert result.code == 200

    result = await client.set_setting(tid=environment, id=data.AUTOSTART_AGENT_MAP, value={"agent1": "", "agent2": "localhost",
                                                                                           "agent3": "user@agent3"})
    assert result.code == 200

    result = await client.set_setting(tid=environment, id=data.AUTOSTART_AGENT_MAP, value="")
    assert result.code == 500


@pytest.mark.asyncio
async def test_clear_environment(client, server, environment):
    """
        Test clearing out an environment
    """
    version = int(time.time())
    result = await client.put_version(tid=environment, version=version, resources=[], unknowns=[], version_info={})
    assert result.code == 200

    result = await client.get_environment(id=environment, versions=10)
    assert result.code == 200
    assert len(result.result["environment"]["versions"]) == 1

    result = await client.clear_environment(id=environment)
    assert result.code == 200

    result = await client.get_environment(id=environment, versions=10)
    assert result.code == 200
    assert len(result.result["environment"]["versions"]) == 0


@pytest.mark.asyncio
async def test_purge_on_delete_requires(client, server, environment):
    """
        Test purge on delete of resources and inversion of requires
    """
    agent = Agent("localhost", {"blah": "localhost"}, environment=environment, code_loader=False)
    await agent.start()
    aclient = agent._client

    version = 1

    resources = [{'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1,path=/tmp/file1],v=%d' % version,
                  'owner': 'root',
                  'path': '/tmp/file1',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'purge_on_delete': True,
                  'version': version},
                 {'group': 'root',
                  'hash': 'b4350bef50c3ec3ee532d4a3f9d6daedec3d2aba',
                  'id': 'std::File[vm2,path=/tmp/file2],v=%d' % version,
                  'owner': 'root',
                  'path': '/tmp/file2',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'purge_on_delete': True,
                  'requires': ['std::File[vm1,path=/tmp/file1],v=%d' % version],
                  'version': version}]

    res = await client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert res.code == 200

    # Release the model and set all resources as deployed
    result = await client.release_version(environment, version, False)
    assert result.code == 200

    now = datetime.now()
    result = await aclient.resource_action_update(environment,
                                                  ['std::File[vm1,path=/tmp/file1],v=%d' % version],
                                                  uuid.uuid4(), "deploy", now, now, "deployed", [], {})
    assert result.code == 200

    result = await aclient.resource_action_update(environment,
                                                  ['std::File[vm2,path=/tmp/file2],v=%d' % version],
                                                  uuid.uuid4(), "deploy", now, now, "deployed", [], {})
    assert result.code == 200

    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["version"] == version
    assert result.result["model"]["total"] == len(resources)
    assert result.result["model"]["done"] == len(resources)
    assert result.result["model"]["released"]
    assert result.result["model"]["result"] == const.VersionState.success.name

    # validate requires and provides
    file1 = [x for x in result.result["resources"] if "file1" in x["id"]][0]
    file2 = [x for x in result.result["resources"] if "file2" in x["id"]][0]

    assert file2["id"] in file1["provides"]
    assert len(file1["attributes"]["requires"]) == 0

    assert len(file2["provides"]) == 0
    assert file1["id"] in file2["attributes"]["requires"]

    result = await client.decomission_environment(id=environment)
    assert result.code == 200

    version = result.result["version"]
    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["total"] == len(resources)

    # validate requires and provides
    file1 = [x for x in result.result["resources"] if "file1" in x["id"]][0]
    file2 = [x for x in result.result["resources"] if "file2" in x["id"]][0]

    assert file2["id"] in file1["attributes"]["requires"]
    assert type(file1["attributes"]["requires"]) == list
    assert len(file1["provides"]) == 0

    assert len(file2["attributes"]["requires"]) == 0
    assert file1["id"] in file2["provides"]
    await agent.stop()


@pytest.mark.asyncio(timeout=20)
async def test_purge_on_delete_compile_failed_with_compile(event_loop, client, server, environment, snippetcompiler):
    config.Config.set("compiler_rest_transport", "request_timeout", "1")

    snippetcompiler.setup_for_snippet("""
    h = std::Host(name="test", os=std::linux)
    f = std::ConfigFile(host=h, path="/etc/motd", content="test", purge_on_delete=true)
    """)
    version, _ = await snippetcompiler.do_export_and_deploy(do_raise=False)

    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["total"] == 1

    snippetcompiler.setup_for_snippet("""
    h = std::Host(name="test")
    """)

    # force deploy by having unknown
    unknown_parameters.append({"parameter": "a", "source": "b"})

    # ensure new version, wait for other second
    await asyncio.sleep(1)

    version, _ = await snippetcompiler.do_export_and_deploy(do_raise=False)
    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["total"] == 0


@pytest.mark.asyncio
async def test_purge_on_delete_compile_failed(client, server, environment):
    """
        Test purge on delete of resources
    """
    agent = Agent("localhost", {"blah": "localhost"}, environment=environment, code_loader=False)
    await agent.start()
    aclient = agent._client

    version = 1

    resources = [{'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1,path=/tmp/file1],v=%d' % version,
                  'owner': 'root',
                  'path': '/tmp/file1',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'purge_on_delete': True,
                  'version': version},
                 {'group': 'root',
                  'hash': 'b4350bef50c3ec3ee532d4a3f9d6daedec3d2aba',
                  'id': 'std::File[vm1,path=/tmp/file2],v=%d' % version,
                  'owner': 'root',
                  'path': '/tmp/file2',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'purge_on_delete': True,
                  'requires': ['std::File[vm1,path=/tmp/file1],v=%d' % version],
                  'version': version},
                 {'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1,path=/tmp/file3],v=%d' % version,
                  'owner': 'root',
                  'path': '/tmp/file3',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'purge_on_delete': True,
                  'version': version}]

    result = await client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # Release the model and set all resources as deployed
    result = await client.release_version(environment, version, False)
    assert result.code == 200

    now = datetime.now()
    result = await aclient.resource_action_update(environment,
                                                  ['std::File[vm1,path=/tmp/file1],v=%d' % version],
                                                  uuid.uuid4(), "deploy", now, now, "deployed", [], {})
    assert result.code == 200

    result = await aclient.resource_action_update(environment,
                                                  ['std::File[vm1,path=/tmp/file2],v=%d' % version],
                                                  uuid.uuid4(), "deploy", now, now, "deployed", [], {})
    assert result.code == 200

    result = await aclient.resource_action_update(environment,
                                                  ['std::File[vm1,path=/tmp/file3],v=%d' % version],
                                                  uuid.uuid4(), "deploy", now, now, "deployed", [], {})
    assert result.code == 200

    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["version"] == version
    assert result.result["model"]["total"] == len(resources)
    assert result.result["model"]["done"] == len(resources)
    assert result.result["model"]["released"]
    assert result.result["model"]["result"] == const.VersionState.success.name

    # New version with only file3
    version = 2
    result = await client.put_version(tid=environment, version=version, resources=[],
                                      unknowns=[{"parameter": "a", "source": "b"}], version_info={const.EXPORT_META_DATA:
                                      {const.META_DATA_COMPILE_STATE: const.Compilestate.failed}})
    assert result.code == 200

    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["total"] == 0
    assert len(result.result["unknowns"]) == 1


@pytest.mark.asyncio
async def test_purge_on_delete(client, server, environment):
    """
        Test purge on delete of resources
    """
    agent = Agent("localhost", {"blah": "localhost"}, environment=environment, code_loader=False)
    await agent.start()
    aclient = agent._client

    version = 1

    resources = [{'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1,path=/tmp/file1],v=%d' % version,
                  'owner': 'root',
                  'path': '/tmp/file1',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'purge_on_delete': True,
                  'version': version},
                 {'group': 'root',
                  'hash': 'b4350bef50c3ec3ee532d4a3f9d6daedec3d2aba',
                  'id': 'std::File[vm1,path=/tmp/file2],v=%d' % version,
                  'owner': 'root',
                  'path': '/tmp/file2',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'purge_on_delete': True,
                  'requires': ['std::File[vm1,path=/tmp/file1],v=%d' % version],
                  'version': version},
                 {'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1,path=/tmp/file3],v=%d' % version,
                  'owner': 'root',
                  'path': '/tmp/file3',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'purge_on_delete': True,
                  'version': version}]

    res = await client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert res.code == 200

    # Release the model and set all resources as deployed
    result = await client.release_version(environment, version, False)
    assert result.code == 200

    now = datetime.now()
    result = await aclient.resource_action_update(environment,
                                                  ['std::File[vm1,path=/tmp/file1],v=%d' % version],
                                                  uuid.uuid4(), "deploy", now, now, "deployed", [], {})
    assert result.code == 200

    result = await aclient.resource_action_update(environment,
                                                  ['std::File[vm1,path=/tmp/file2],v=%d' % version],
                                                  uuid.uuid4(), "deploy", now, now, "deployed", [], {})
    assert result.code == 200

    result = await aclient.resource_action_update(environment,
                                                  ['std::File[vm1,path=/tmp/file3],v=%d' % version],
                                                  uuid.uuid4(), "deploy", now, now, "deployed", [], {})
    assert result.code == 200

    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["version"] == version
    assert result.result["model"]["total"] == len(resources)
    assert result.result["model"]["done"] == len(resources)
    assert result.result["model"]["released"]
    assert result.result["model"]["result"] == const.VersionState.success.name

    # New version with only file3
    version = 2
    res3 = {'group': 'root',
            'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
            'id': 'std::File[vm1,path=/tmp/file3],v=%d' % version,
            'owner': 'root',
            'path': '/tmp/file3',
            'permissions': 644,
            'purged': False,
            'reload': False,
            'requires': [],
            'purge_on_delete': True,
            'version': version}
    result = await client.put_version(tid=environment, version=version, resources=[res3], unknowns=[], version_info={})
    assert result.code == 200

    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["total"] == 3

    # validate requires and provides
    file1 = [x for x in result.result["resources"] if "file1" in x["id"]][0]
    file2 = [x for x in result.result["resources"] if "file2" in x["id"]][0]
    file3 = [x for x in result.result["resources"] if "file3" in x["id"]][0]

    assert file1["attributes"]["purged"]
    assert file2["attributes"]["purged"]
    assert not file3["attributes"]["purged"]
    await agent.stop()


@pytest.mark.asyncio
async def test_purge_on_delete_ignore(client, server, environment):
    """
        Test purge on delete behavior for resources that have not longer purged_on_delete set
    """
    agent = Agent("localhost", {"blah": "localhost"}, environment=environment, code_loader=False)
    await agent.start()
    aclient = agent._client

    # Version 1 with purge_on_delete true
    version = 1

    resources = [{'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1,path=/tmp/file1],v=%d' % version,
                  'owner': 'root',
                  'path': '/tmp/file1',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'purge_on_delete': True,
                  'version': version}]

    res = await client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert res.code == 200

    # Release the model and set all resources as deployed
    result = await client.release_version(environment, version, False)
    assert result.code == 200

    now = datetime.now()
    result = await aclient.resource_action_update(environment,
                                                  ['std::File[vm1,path=/tmp/file1],v=%d' % version],
                                                  uuid.uuid4(), "deploy", now, now, "deployed", [], {})
    assert result.code == 200

    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["version"] == version
    assert result.result["model"]["total"] == len(resources)
    assert result.result["model"]["done"] == len(resources)
    assert result.result["model"]["released"]
    assert result.result["model"]["result"] == const.VersionState.success.name

    # Version 2 with purge_on_delete false
    version = 2

    resources = [{'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1,path=/tmp/file1],v=%d' % version,
                  'owner': 'root',
                  'path': '/tmp/file1',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'purge_on_delete': False,
                  'version': version}]

    res = await client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert res.code == 200

    # Release the model and set all resources as deployed
    result = await client.release_version(environment, version, False)
    assert result.code == 200

    now = datetime.now()
    result = await aclient.resource_action_update(environment,
                                                  ['std::File[vm1,path=/tmp/file1],v=%d' % version],
                                                  uuid.uuid4(), "deploy", now, now, "deployed", [], {})
    assert result.code == 200

    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["version"] == version
    assert result.result["model"]["total"] == len(resources)
    assert result.result["model"]["done"] == len(resources)
    assert result.result["model"]["released"]
    assert result.result["model"]["result"] == const.VersionState.success.name

    # Version 3 with no resources
    version = 3
    resources = []
    res = await client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert res.code == 200

    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["version"] == version
    assert result.result["model"]["total"] == len(resources)
    await agent.stop()


@pytest.mark.asyncio
async def test_tokens(server_multi, client_multi, environment):
    # Test using API tokens
    test_token = client_multi._transport_instance.token
    token = await client_multi.create_token(environment, ["api"], idempotent=True)
    jot = token.result["token"]

    assert jot != test_token

    client_multi._transport_instance.token = jot

    # try to access a non environment call (global)
    result = await client_multi.list_environments()
    assert result.code == 401

    result = await client_multi.list_versions(environment)
    assert result.code == 200

    token = await client_multi.create_token(environment, ["agent"], idempotent=True)
    agent_jot = token.result["token"]

    client_multi._transport_instance.token = agent_jot
    result = await client_multi.list_versions(environment)
    assert result.code == 401


def make_source(collector, filename, module, source, req):
    myhash = hash_file(source.encode())
    collector[myhash] = [filename, module, source, req]
    return collector


@pytest.mark.asyncio(timeout=30)
async def test_code_upload(motor, server_multi, client_multi, agent_multi, environment_multi):
    """
        Test the server to manage the updates on a model during agent deploy
    """
    version = 1

    resources = [{'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1.dev.inmanta.com,path=/etc/sysconfig/network],v=%d' % version,
                  'owner': 'root',
                  'path': '/etc/sysconfig/network',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'version': version}]

    res = await client_multi.put_version(
        tid=environment_multi, version=version, resources=resources, unknowns=[], version_info={}
    )
    assert res.code == 200

    sources = make_source({}, "a.py", "std.test", "wlkvsdbhewvsbk vbLKBVWE wevbhbwhBH", [])
    sources = make_source(sources, "b.py", "std.xxx", "rvvWBVWHUvejIVJE UWEBVKW", ["pytest"])

    res = await client_multi.upload_code(tid=environment_multi, id=version, resource="std::File", sources=sources)
    assert res.code == 200

    res = await agent_multi._client.get_code(tid=environment_multi, id=version, resource="std::File")
    assert res.code == 200
    assert res.result["sources"] == sources


@pytest.mark.asyncio(timeout=30)
async def test_batched_code_upload(motor, server_multi, client_multi, sync_client_multi, environment_multi, agent_multi):
    """
        Test the server to manage the updates on a model during agent deploy
    """
    version = 1

    resources = [{'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1.dev.inmanta.com,path=/etc/sysconfig/network],v=%d' % version,
                  'owner': 'root',
                  'path': '/etc/sysconfig/network',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'version': version}]

    res = await client_multi.put_version(
        tid=environment_multi, version=version, resources=resources, unknowns=[], version_info={}
    )
    assert res.code == 200

    asources = make_source({}, "a.py", "std.test", "wlkvsdbhewvsbk vbLKBVWE wevbhbwhBH", [])
    asources = make_source(asources, "b.py", "std.xxx", "rvvWBVWHUvejIVJE UWEBVKW", ["pytest"])

    bsources = make_source({}, "a.py", "std.test", "wlkvsdbhewvsbk vbLKBVWE wevbhbwhBH", [])
    bsources = make_source(bsources, "c.py", "std.xxx", "enhkahEUWLGBVFEHJ UWEBVKW", ["pytest"])

    csources = make_source({}, "a.py", "sss", "ujekncedsiekvsd", [])

    sources = {"std::File": asources,
               "std::Other": bsources,
               "std:xxx": csources
               }

    await asyncio.get_event_loop().run_in_executor(
        None, lambda: upload_code(sync_client_multi, environment_multi, version, sources)
    )

    for name, sourcemap in sources.items():
        res = await agent_multi._client.get_code(tid=environment_multi, id=version, resource=name)
        assert res.code == 200
        assert res.result["sources"] == sourcemap


@pytest.mark.asyncio(timeout=30)
async def test_legacy_code(motor, server_multi, client_multi, environment_multi, agent_multi):
    """
        Test the server to manage the updates on a model during agent deploy
    """
    version = 2

    resources = [{'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1.dev.inmanta.com,path=/etc/sysconfig/network],v=%d' % version,
                  'owner': 'root',
                  'path': '/etc/sysconfig/network',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'version': version}]

    res = await client_multi.put_version(
        tid=environment_multi, version=version, resources=resources, unknowns=[], version_info={}
    )
    assert res.code == 200

    sources = {"a.py": "ujeknceds", "b.py": "weknewbevbvebedsvb"}

    code = data.Code(environment=UUID(environment_multi), version=version, resource="std::File", sources=sources)
    await code.insert()

    res = await agent_multi._client.get_code(tid=environment_multi, id=version, resource="std::File")
    assert res.code == 200
    assert res.result["sources"] == sources


@pytest.mark.asyncio(timeout=30)
async def test_resource_action_log(motor, server_multi, client_multi, environment):
    version = 1
    resources = [{'group': 'root',
                  'hash': '89bf880a0dc5ffc1156c8d958b4960971370ee6a',
                  'id': 'std::File[vm1.dev.inmanta.com,path=/etc/sysconfig/network],v=%d' % version,
                  'owner': 'root',
                  'path': '/etc/sysconfig/network',
                  'permissions': 644,
                  'purged': False,
                  'reload': False,
                  'requires': [],
                  'version': version}]
    res = await client_multi.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert res.code == 200

    resource_action_log = server.Server.get_resource_action_log_file(environment)
    assert os.path.isfile(resource_action_log)
    assert os.stat(resource_action_log).st_size != 0


@pytest.mark.asyncio(timeout=30)
async def test_invalid_sid(server_multi, client_multi, environment_multi):
    """
        Test the server to manage the updates on a model during agent deploy
    """
    # request get_code with a compiler client that does not have a sid
    res = await client_multi.get_code(tid=environment_multi, id=1, resource="std::File")
    assert res.code == 400
    assert res.result["message"] == "Invalid request: this is an agent to server call, it should contain an agent session id"



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import base64
from urllib import request, error
from collections import defaultdict
from configparser import ConfigParser, Interpolation
import json
import logging
import os
import re
import sys
import uuid
import warnings
import ssl

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers
from inmanta import const

from typing import Optional, Callable, TypeVar, Generic, Dict, List, cast, Union

LOGGER = logging.getLogger(__name__)


def _normalize_name(name: str) -> str:
    return name.replace("_", "-")


class LenientConfigParser(ConfigParser):

    def optionxform(self, name: str) -> str:
        name = _normalize_name(name)
        return super(LenientConfigParser, self).optionxform(name)


class Config(object):
    __instance: Optional["Config"] = None
    __config_definition: Dict[str, Dict[str, "Option"]] = defaultdict(lambda: {})

    @classmethod
    def get_config_options(cls) -> Dict[str, Dict[str, str]]:
        return cls.__config_definition

    @classmethod
    def load_config(cls, config_file: str = None) -> None:
        """
        Load the configuration file
        """
        config = LenientConfigParser(interpolation=Interpolation())

        files = ["/etc/inmanta.cfg", os.path.expanduser("~/.inmanta.cfg"), ".inmanta",
                 ".inmanta.cfg"]
        if config_file is not None:
            files.append(config_file)

        config.read(files)
        cls.__instance = config

    @classmethod
    def _get_instance(cls) -> "Config":
        if cls.__instance is None:
            raise Exception("Load the configuration first")

        return cls.__instance

    @classmethod
    def _reset(cls) -> None:
        cls.__instance = None

    # noinspection PyNoneFunctionAssignment
    @classmethod
    def get(cls, section: Optional[str] = None, name: Optional[str] = None, default_value: Optional[str] = None) -> str:
        """
            Get the entire compiler or get a value directly
        """
        cfg = cls._get_instance()
        if section is None:
            return cfg
        name = _normalize_name(name)

        opt = cls.validate_option_request(section, name, default_value)

        val = cfg.get(section, name, fallback=default_value)
        if not opt:
            return val
        return opt.validate(val)

    @classmethod
    def getboolean(cls, section: str, name: str, default_value: str = None) -> bool:
        """
            Return a boolean from the configuration
        """
        cls.validate_option_request(section, name, default_value)
        return cls._get_instance().getboolean(section, name, fallback=default_value)

    @classmethod
    def set(cls, section: str, name: str, value: str) -> None:
        """
            Override a value
        """
        name = _normalize_name(name)

        if section not in cls._get_instance():
            cls._get_instance().add_section(section)
        cls._get_instance().set(section, name, value)

    @classmethod
    def register_option(cls, option: "Option") -> None:
        cls.__config_definition[option.section][option.name] = option

    @classmethod
    def validate_option_request(cls, section: str, name: str, default_value: str) -> Optional["Option"]:
        if section not in cls.__config_definition:
            LOGGER.warning("Config section %s not defined" % (section))
            # raise Exception("Config section %s not defined" % (section))
            return
        if name not in cls.__config_definition[section]:
            LOGGER.warning("Config name %s not defined in section %s" % (name, section))
            # raise Exception("Config name %s not defined in section %s" % (name, section))
            return
        opt = cls.__config_definition[section][name]
        if default_value is not None and opt.get_default_value() != default_value:
            LOGGER.warning("Inconsistent default value for option %s.%s: defined as %s, got %s" %
                        (section, name, opt.default, default_value))

        return opt


def is_int(value: str) -> int:
    """int"""
    return int(value)


def is_time(value: str) -> int:
    """time"""
    return int(value)


def is_bool(value: str) -> bool:
    """bool"""
    if type(value) == bool:
        return cast(bool, value)
    return Config._get_instance()._convert_to_boolean(value)


def is_list(value: str) -> List[str]:
    """list"""
    return [x.strip() for x in value.split(",")]


def is_map(map_in: str) -> Dict[str, str]:
    """map"""
    map_out = {}
    if map_in is not None:
        mappings = map_in.split(",")

        for mapping in mappings:
            parts = re.split("=", mapping.strip(), 1)
            if len(parts) == 2:
                key = parts[0].strip()
                value = parts[1].strip()
                if key != "" and value != "":
                    map_out[key] = value

    return map_out


def is_str(value: str) -> str:
    """str"""
    return str(value)


def is_str_opt(value: str) -> Optional[str]:
    """optional str"""
    if value is None:
        return None
    return str(value)


def is_uuid_opt(value: str) -> uuid.UUID:
    """optional uuid"""
    if value is None:
        return None
    return uuid.UUID(value)


T = TypeVar("T")

class Option(Generic[T]):
    """
    Defines an option and exposes it for use

    All config option should be define prior to use
    For the document generator to work properly, they should be defined at the module level.

    :param section: section in the config file
    :param name: name of the option
    :param default: default value for this option
        the default value is either a value or a function.
        If it is a value, `str(default)` will be used a default value.
        If it is a function, its doc string will be used to represent the value in documentation.
        and its return value as the actual default value
    :param documentation: the documentation for this option
    :param validator: a function responsible for turning the string representation of the option into the correct type.
        Its docstring is used as representation for the type of the option.
    :param predecessor_option: The Option that was deprecated in favour of this option.
    """

    def __init__(
        self,
        section: str,
        name: str,
        default: Union[str, bool, int, None, Callable],
        documentation: str,
        validator: Callable[[str], T] = is_str,
        predecessor_option: "Option" = None
    ) -> None:
        self.section = section
        self.name = _normalize_name(name)
        self.validator = validator
        self.documentation = documentation
        self.default = default
        self.predecessor_option = predecessor_option
        Config.register_option(self)

    def get(self) -> T:
        cfg = Config._get_instance()
        if self.predecessor_option:
            has_deprecated_option = cfg.has_option(self.predecessor_option.section, self.predecessor_option.name)
            has_new_option = cfg.has_option(self.section, self.name)
            if has_deprecated_option and not has_new_option:
                warnings.warn("Config option %s is deprecated. Use %s instead." % (self.predecessor_option.name, self.name),
                              category=DeprecationWarning)
                return self.predecessor_option.get()
        out = cfg.get(self.section, self.name, fallback=self.get_default_value())
        return self.validate(out)

    def get_type(self) -> Optional[str]:
        if callable(self.validator):
            return self.validator.__doc__
        return None

    def get_default_desc(self) -> str:
        defa = self.default
        if callable(defa):
            return "$%s" % defa.__doc__
        else:
            return defa

    def validate(self, value: str) -> T:
        return self.validator(value)

    def get_default_value(self):
        defa = self.default
        if callable(defa):
            return defa()
        else:
            return defa

    def set(self, value: str) -> None:
        """ Only for tests"""
        cfg = Config._get_instance()
        cfg.set(self.section, self.name, value)

#############################
# Config
#
# Global config options are defined here
#############################
# flake8: noqa: H904
state_dir = Option("config", "state_dir", "/var/lib/inmanta",
                   "The directory where the server stores its state")

log_dir = Option("config", "log_dir", "/var/log/inmanta",
                 "The directory where the server stores log file. Currently this is only for the output of embedded agents.")


def get_executable():
    """os.path.abspath(sys.argv[0]) """
    try:
        return os.path.abspath(sys.argv[0])
    except:
        return None


def get_default_nodename():
    """ socket.gethostname() """
    import socket
    return socket.gethostname()


nodename = Option("config", "node-name", get_default_nodename,
                  "Force the hostname of this machine to a specific value", is_str)


###############################
# Transport Config
###############################
class TransportConfig(object):
    """
        A class to register the config options for Client classes
    """
    def __init__(self, name: str, port: int=8888) -> None:
        self.prefix = "%s_rest_transport" % name
        self.host = Option(self.prefix, "host", "localhost", "IP address or hostname of the server", is_str)
        self.port = Option(self.prefix, "port", port, "Server port", is_int)
        self.ssl = Option(self.prefix, "ssl", False, "Connect using SSL?", is_bool)
        self.ssl_ca_cert_file = Option(self.prefix, "ssl_ca_cert_file", None,
                                       "CA cert file used to validate the server certificate against", is_str_opt)
        self.token = Option(self.prefix, "token", None, "The bearer token to use to connect to the API", is_str_opt)
        self.request_timeout = Option(self.prefix, "request_timeout", 120, "The time before a request times out in seconds", is_int)


compiler_transport = TransportConfig("compiler")
TransportConfig("client")
cmdline_rest_transport = TransportConfig("cmdline")
# LCM support should move to a server extension
service_api_transport = TransportConfig("service_api", port=8889)


#############################
# auth
#############################
AUTH_JWT_PREFIX = "auth_jwt_"

class AuthJWTConfig(object):
    """
        Auth JWT configuration manager
    """
    sections: Dict[str, "AuthJWTConfig"] = {}
    issuers: Dict[str, "AuthJWTConfig"] = {}

    validate_cert: bool

    @classmethod
    def list(cls) -> List[str]:
        """
            Return a list of all defined auth jwt configurations. This method will load new sections if they were added
            since the last invocation.
        """
        cfg = Config._get_instance()
        prefix_len = len(AUTH_JWT_PREFIX)

        for section in cfg.keys():
            if section[:prefix_len] == AUTH_JWT_PREFIX:
                name = section[prefix_len:]
                if name not in cls.sections:
                    obj = cls(name, section, cfg[section])
                    cls.sections[name] = obj
                    if obj.issuer in cls.issuers:
                        raise ValueError("Only oner configuration per issuer is supported")

                    cls.issuers[obj.issuer] = obj

        # Verify that only one has sign set to true
        sign = False
        for name, cfg in cls.sections.items():
            if cfg.sign:
                if sign:
                    raise ValueError("Only one auth_jwt section may have sign set to true")
                else:
                    sign = True

        if len(cls.sections.keys()) > 0 and not sign:
            raise ValueError("One auth_jwt section should have sign set to true")

        return list(cls.sections.keys())

    @classmethod
    def get(cls, name: str) -> Optional["AuthJWTConfig"]:
        """
            Get the config with the given name
        """
        cls.list()
        if name in cls.sections:
            return cls.sections[name]
        return None

    @classmethod
    def get_sign_config(cls) -> Optional["AuthJWTConfig"]:
        """
            Get the configuration with sign is true
        """
        cls.list()
        for cfg in cls.sections.values():
            if cfg.sign:
                return cfg

    @classmethod
    def get_issuer(cls, issuer: str) -> Optional["AuthJWTConfig"]:
        """
            Get the config for the given issuer. Only when no auth config has been loaded yet, the configuration will be loaded
            again. For loading additional configuration, call list() first. This method is in the auth path for each API
            request.
        """
        if len(cls.issuers) == 0:
            cls.list()
        if issuer in cls.issuers:
            return cls.issuers[issuer]
        return None

    def __init__(self, name: str, section: str, config: Config):
        self.name = name
        self.section = section
        self._config = config
        if "algorithm" not in config:
            raise ValueError("algorithm is required in %s section" % self.section)

        self.algo = config["algorithm"]
        self.validate_generic()

        if self.algo.lower() == "hs256":
            self.validate_hs265()
        elif self.algo.lower() == "rs256":
            self.validate_rs265()
        else:
            raise ValueError("Algorithm %s in %s is not support " % (self.algo, self.section))

    def validate_generic(self) -> None:
        """
            Validate  and parse the generic options that are valid for all algorithms
        """
        if "sign" in self._config:
            self.sign = is_bool(self._config["sign"])
        else:
            self.sign = False

        if "client_types" not in self._config:
            raise ValueError("client_types is a required options for %s" % self.section)

        self.client_types = is_list(self._config["client_types"])
        for ct in self.client_types:
            if ct not in const.VALID_CLIENT_TYPES:
                raise ValueError("invalid client_type %s in %s" % (ct, self.section))

        if "expire" in self._config:
            self.expire = is_int(self._config["expire"])
        else:
            self.expire = 0

        if "issuer" in self._config:
            self.issuer = is_str(self._config["issuer"])
        else:
            self.issuer = "https://localhost:8888/"

        if "audience" in self._config:
            self.audience = is_str(self._config["audience"])
        else:
            self.audience = self.issuer

    def validate_hs265(self) -> None:
        """
            Validate and parse HS256 algorithm configuration
        """
        if "key" not in self._config:
            raise ValueError("key is required in %s for algorithm %s" % (self.section, self.algo))

        self.key = base64.urlsafe_b64decode((self._config["key"] + "==").encode("ascii"))
        if len(self.key) < 32:
            raise ValueError("HS256 requires a key of 32 bytes (256 bits) or longer in " + self.section)

    def _load_public_key(self, e: str, n: str) -> str:
        def to_int(x: str) -> int:
            bs = base64.urlsafe_b64decode(x + "==")
            return int.from_bytes(bs, byteorder="big")

        ei = to_int(e)
        ni = to_int(n)
        numbers = RSAPublicNumbers(ei, ni)
        public_key = numbers.public_key(backend=default_backend())
        pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        return pem

    def validate_rs265(self) -> None:
        """
            Validate and parse RS256 algorithm configuration
        """
        if "jwks_uri" not in self._config:
            raise ValueError("jwks_uri is required for RS256 based providers in section %s" % self.section)

        self.jwks_uri = self._config["jwks_uri"]

        if "validate_cert" in self._config:
            self.validate_cert = self._config.getboolean("validate_cert")
        else:
            self.validate_cert = True

        ctx = None
        if not self.validate_cert:
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE

        try:
            with request.urlopen(self.jwks_uri, context=ctx) as response:
                key_data = json.loads(response.read().decode('utf-8'))
        except error.URLError as e:
            # HTTPError is raised for non-200 responses; the response
            # can be found in e.response.
            raise ValueError("Unable to load key data for %s using the provided jwks_uri. Got error: %s" %
                             (self.section, e.response))
        except Exception as e:
            # Other errors are possible, such as IOError.
            raise ValueError("Unable to load key data for %s using the provided jwks_uri." % (self.section))

        self.keys = {}
        for key in key_data["keys"]:
            self.keys[key["kid"]] = self._load_public_key(key["e"], key["n"])



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import pytest
import inmanta.compiler as compiler
from inmanta.execute.proxy import DynamicProxy


def proxy_object(snippetcompiler, snippet, var):
    snippetcompiler.setup_for_snippet(snippet)
    (_, root) = compiler.do_compile()

    scope = root.get_child("__config__").scope

    value = scope.lookup(var).get_value()
    return DynamicProxy.return_value(value)


def test_basic_object(snippetcompiler):
    px = proxy_object(snippetcompiler, """
entity Test1:
string x
end
implement Test1 using std::none

a = Test1(x="a")
""", "a")

    assert px.x == "a"
    with pytest.raises(Exception):
        px.x = "b"


def test_dict_attr(snippetcompiler):
    px = proxy_object(snippetcompiler, """
entity Test1:
dict x = {}
end
implement Test1 using std::none

a = Test1(x={"a":"A"})
""", "a")

    dic = px.x
    assert dic["a"] == "A"
    with pytest.raises(Exception):
        dic["a"] = "Z"

    with pytest.raises(Exception):
        dic["b"]

    with pytest.raises(Exception):
        dic["b"] = "a"



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
import os
import random
import subprocess
import sys
import time
import socket

from inmanta import module, config, protocol, const, data, mongoproc


LOGGER = logging.getLogger(__name__)
PORT_START = 40000
MAX_TRIES = 25


cfg_prj = config.Option("deploy", "project", "deploy", "The project name to use in the deploy", config.is_str_opt)
cfg_env = config.Option("deploy", "environment", "deploy", "The environment name to use in the deploy", config.is_str_opt)


class FinishedException(Exception):
    """
        This exception is raised when the deploy is ready
    """


class Deploy(object):
    def __init__(self, mongoport=0):
        self._mongoproc = None
        self._mongoport = mongoport
        self._server_port = 0
        self._data_path = None
        self._server_proc = None
        self._agent = None
        self._client = None

        self._environment_id = None

        self._agent_ready = False

        loud_logger = logging.getLogger("inmanta.protocol")
        loud_logger.propagate = False

        loud_logger = logging.getLogger("tornado")
        loud_logger.propagate = False

    def _ensure_dir(self, path):
        if not os.path.exists(path):
            LOGGER.debug("Creating directory %s", path)
            os.mkdir(path)

    def setup_server(self, no_agent_log):
        state_dir = os.path.join(self._data_path, "state")
        self._ensure_dir(state_dir)

        log_dir = os.path.join(self._data_path, "logs")
        self._ensure_dir(log_dir)

        self._server_port = PORT_START + random.randint(0, 2000)
        assert self._server_port != self._mongoport

        config.Config.set("client_rest_transport", "port", str(self._server_port))

        config_file = """
[database]
name=inmanta
port=%(mongo_port)s

[config]
state-dir=%(state_dir)s
log-dir=%(log_dir)s

[server_rest_transport]
port=%(server_port)s

[agent_rest_transport]
port=%(server_port)s

[compiler_rest_transport]
port=%(server_port)s

[client_rest_transport]
port=%(server_port)s

[cmdline_rest_transport]
port=%(server_port)s
""" % {"state_dir": state_dir, "log_dir": log_dir, "server_port": self._server_port, "mongo_port": self._mongoport}

        server_config = os.path.join(self._data_path, "server.cfg")
        with open(server_config, "w+") as fd:
            fd.write(config_file)

        args = [sys.executable, "-m", "inmanta.app", "-vvv", "-c", server_config, "server"]

        self._server_proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        LOGGER.debug("Started server on port %d", self._server_port)

        while self._server_proc.poll() is None:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                try:
                    s.connect(('localhost', int(self._server_port)))
                    return True
                except (IOError, socket.error):
                    time.sleep(0.25)
            finally:
                s.close()

        return False

    def setup_mongodb(self):
        # start a local mongodb on a random port
        if self._mongoport == 0:
            log_dir = os.path.join(self._data_path, "logs")
            self._ensure_dir(log_dir)

            self._mongoport = PORT_START + random.randint(0, 2000)
            mongo_dir = os.path.join(self._data_path, "mongo")
            self._mongoproc = mongoproc.MongoProc(db_path=mongo_dir, port=self._mongoport,
                                                  log_path=os.path.join(log_dir, "mongod.log"))
            LOGGER.debug("Starting mongodb on port %d", self._mongoport)
            if not self._mongoproc.start():
                LOGGER.error("Unable to start mongodb instance on port %d and data directory %s", self._mongoport, mongo_dir)
                return False

        return True

    def _create_project(self, project_name):
        LOGGER.debug("Creating project %s", project_name)
        result = self._client.create_project(project_name)
        if result.code != 200:
            LOGGER.error("Unable to create project %s", project_name)
            return False

        return result.result["project"]["id"]

    def _create_environment(self, project_id, environment_name):
        LOGGER.debug("Creating environment %s in project %s", environment_name, project_id)
        result = self._client.create_environment(project_id=project_id, name=environment_name)
        if result.code != 200:
            LOGGER.error("Unable to create environment %s", environment_name)
            return False

        return result.result["environment"]["id"]

    def _latest_version(self, environment_id):
        result = self._client.list_versions(tid=environment_id)
        if result.code != 200:
            LOGGER.error("Unable to get all version of environment %s", environment_id)
            return None

        if "versions" in result.result and len(result.result["versions"]) > 0:
            versions = [x["version"] for x in result.result["versions"]]
            sorted(versions)
            return versions[0]

        return None

    def setup_project(self):
        """
            Set up the configured project and environment on the embedded server
        """
        self._client = protocol.SyncClient("client")

        # get config
        project_name = cfg_prj.get()
        if project_name is None:
            LOGGER.error("The name of the project should be configured for an all-in-one deploy")
            return False

        environment_name = cfg_env.get()
        if environment_name is None:
            LOGGER.error("The name of the environment in the project should be configured for an all-in-one deploy")
            return False

        # wait and check to see if the server is up
        tries = 0
        while tries < MAX_TRIES:
            try:
                self._client.list_projects()
                break
            except Exception:
                tries += 1

        # get project id
        projects = self._client.list_projects()
        if projects.code != 200:
            LOGGER.error("Unable to retrieve project listing from the server")
            return False

        project_id = None
        for project in projects.result["projects"]:
            if project_name == project["name"]:
                project_id = project["id"]
                break

        if project_id is None:
            project_id = self._create_project(project_name)
            if not project_id:
                return False

        # get or create the environment
        environments = self._client.list_environments()
        if environments.code != 200:
            LOGGER.error("Unable to retrieve environments from server")
            return False

        for env in environments.result["environments"]:
            if project_id == env["project"] and environment_name == env["name"]:
                self._environment_id = env["id"]
                break

        if self._environment_id is None:
            self._environment_id = self._create_environment(project_id, environment_name)
            if not self._environment_id:
                return False

        return True

    def setup(self, no_agent_log=True):
        """
            Run inmanta locally
        """
        # create local storage
        project = module.Project.get()
        self._data_path = os.path.join(project.project_path, "data", "deploy")
        LOGGER.debug("Storing state data in %s", self._data_path)
        self._ensure_dir(os.path.join(project.project_path, "data"))
        self._ensure_dir(self._data_path)

        if not self.setup_mongodb():
            return False

        if not self.setup_server(no_agent_log):
            return False

        if not self.setup_project():
            LOGGER.error("Failed to setup project")
            return False

        return True

    def export(self):
        """
            Export a version to the embedded server
        """
        inmanta_path = [sys.executable, "-m", "inmanta.app"]

        cmd = inmanta_path + ["-vvv", "export", "-e", str(self._environment_id), "--server_address", "localhost",
                              "--server_port", str(self._server_port)]

        sub_process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        log_out, log_err = sub_process.communicate()

        if sub_process.returncode > 0:
            print("An error occurred while compiling the model:")
            if len(log_out) > 0:
                print(log_out.decode())
            if len(log_err) > 0:
                print(log_err.decode())

            return False

        LOGGER.info("Export of model complete")
        return True

    def get_agents_for_model(self, version):
        version_result = self._client.get_version(tid=self._environment_id, id=version)
        if version_result.code != 200:
            LOGGER.error("Unable to get version %d of environment %s", version, self._environment_id)
            return []

        return {x["agent"] for x in version_result.result["resources"]}

    def deploy(self, dry_run):
        version = self._latest_version(self._environment_id)
        LOGGER.info("Latest version for created environment is %s", version)
        if version is None:
            return

        # Update the agentmap to autostart all agents
        agents = self.get_agents_for_model(version)
        LOGGER.debug("Agent(s) %s defined, adding them to autostart agent map", ", ".join(agents))
        result = self._client.get_setting(tid=self._environment_id, id=data.AUTOSTART_AGENT_MAP)

        # release the version!
        if not dry_run:
            self._client.release_version(tid=self._environment_id, id=version, push=True,
                                         agent_trigger_method=const.AgentTriggerMethod.push_full_deploy)
            self.progress_deploy_report(version)

        else:
            result = self._client.dryrun_request(tid=self._environment_id, id=version)
            dryrun_id = result.result["dryrun"]["id"]
            self.progress_dryrun_report(dryrun_id)

    def _get_deploy_stats(self, version):
        version_result = self._client.get_version(tid=self._environment_id, id=version)
        if version_result.code != 200:
            LOGGER.error("Unable to get version %d of environment %s", version, self._environment_id)
            return []

        total = 0
        deployed = 0
        ready = {}

        for res in version_result.result["resources"]:
            total += 1
            if res["status"] != const.ResourceState.available.name:
                deployed += 1
                ready[res["id"]] = res["status"]

        return total, deployed, ready

    def progress_deploy_report(self, version):
        print("Starting deploy")
        current_ready = set()
        total = 0
        deployed = -1
        while total > deployed:
            total, deployed, ready = self._get_deploy_stats(version)

            new = ready.keys() - current_ready
            current_ready = ready

            # if we already printed progress, move cursor one line up
            if deployed >= 0:
                sys.stdout.write('\033[1A')
                sys.stdout.flush()

            for res in new:
                print("%s - %s" % (res, ready[res]))

            print("[%d / %d]" % (deployed, total))
            time.sleep(1)

        print("Deploy ready")

    def _get_dryrun_status(self, dryrun_id):
        result = self._client.dryrun_report(self._environment_id, dryrun_id)

        if result.code != 200:
            raise Exception("Unable to get dryrun report")

        data = result.result["dryrun"]
        return data["total"], data["todo"], data["resources"]

    def progress_dryrun_report(self, dryrun_id):
        print("Starting dryrun")

        current_ready = set()
        total = 0
        todo = 1
        while todo > 0:
            # if we already printed progress, move cursor one line up
            if len(current_ready) > 0:
                sys.stdout.write('\033[1A')
                sys.stdout.flush()

            total, todo, ready = self._get_dryrun_status(dryrun_id)

            new = ready.keys() - current_ready
            current_ready = ready

            for res in new:
                changes = ready[res]["changes"]
                if len(changes) == 0:
                    print("%s - no changes" % res)
                else:
                    print("%s:" % res)
                    for field, values in changes.items():
                        if field == "hash":
                            diff_result = self._client.diff(a=values[0], b=values[1])
                            if diff_result.code == 200:
                                print("  - content:")
                                diff_value = diff_result.result
                                print("    " + "    ".join(diff_value["diff"]))
                        else:
                            print("  - %s:" % field)
                            print("    from: %s" % values[0])
                            print("    to:   %s" % values[1])

                        print("")

            print("[%d / %d]" % (total - todo, total))
            time.sleep(1)

        raise FinishedException()

    def run(self, options, only_setup=False):
        self.export()
        self.deploy(dry_run=options.dryrun)

    def stop(self):
        if self._server_proc is not None:
            self._server_proc.kill()

        if self._mongoproc is not None:
            self._mongoproc.stop()



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import pytest

from inmanta.ast import AttributeException
from inmanta.ast import TypingException
import inmanta.compiler as compiler


def test_lnr_on_double_is_defined(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    string? two
end

Test.one [0:1] -- Test

implement Test using std::none when self.one.two is defined

a = Test(two="b")
a.one = a
"""
    )
    compiler.do_compile()


def test_double_define(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    string test
    string? test
    bool test
end
"""
    )
    with pytest.raises(TypingException):
        compiler.do_compile()


def test_536_number_cast(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Network:
    number segmentation_id
end
implement Network using std::none
net1 = Network(segmentation_id="10")
"""
    )
    with pytest.raises(AttributeException):
        compiler.do_compile()



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: bart@inmanta.com
"""
import uuid
import datetime
import time

from inmanta import data
from inmanta import const
import pytest
import pymongo
import logging


class Doc(data.BaseDocument):
    name = data.Field(field_type=str, required=True)
    field1 = data.Field(field_type=str, default=None)
    field2 = data.Field(field_type=bool, default=False)
    field3 = data.Field(field_type=list, default=[])
    field4 = data.Field(field_type=dict, default={})

    __indexes__ = [
        dict(keys=[("name", pymongo.ASCENDING)], unique=True)
    ]


class EnumDoc(data.BaseDocument):
    action = data.Field(field_type=const.ResourceAction, required=True)


@pytest.mark.asyncio
async def test_motor(motor):
    await motor.testCollection.insert_one({"a": 1, "b": "abcd"})
    results = await motor.testCollection.find_one({"a": {"$gt": 0}})

    assert "_id" in results

    await motor.testCollection.insert_one({"a": {"b": {"c": 1}}})
    results = motor.testCollection.find({})

    while (await results.fetch_next):
        assert "a" in results.next_object()


def test_collection_naming():
    assert Doc.collection_name() == "Doc"


def test_ctor():
    # Mongo data should not contain an id
    with pytest.raises(AttributeError):
        Doc(from_mongo=True, id="1234")

    # ID is autogenerated for new objects
    with pytest.raises(AttributeError):
        Doc(from_mongo=True, id="1234")

    # Required attributes have to be set in the ctor
    with pytest.raises(AttributeError):
        Doc()

    # Only defined fields are allowed
    with pytest.raises(AttributeError):
        Doc(test="")

    # A required field cannot be none
    with pytest.raises(TypeError):
        Doc(name=None)

    # Fields require the correct type
    with pytest.raises(TypeError):
        Doc(name=1234)


def test_document_def():
    t = Doc(name="doc")
    with pytest.raises(TypeError):
        t.id = "1234"

    t.id = uuid.uuid4()
    json = t.to_dict()

    assert "id" in json
    assert "_id" in t.to_mongo()


@pytest.mark.asyncio
async def test_document_insert(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")

    await d.insert()
    docs = await motor.Doc.find({}).to_list(length=10)
    assert len(docs) == 1

    doc = await Doc.get_by_id(d.id)
    assert doc.name == d.name


@pytest.mark.asyncio
async def test_get_by_id(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    await d.insert()

    d1 = await Doc.get_by_id(d.id)
    assert d1.name == d.name

    d2 = await Doc.get_by_id(uuid.uuid4())
    assert d2 is None


@pytest.mark.asyncio
async def test_defaults(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    await d.insert()

    assert d.to_dict()["field1"] is None
    assert not d.field2

    d2 = await Doc.get_by_id(d.id)
    d2.insert()

    assert not d2.field2

    d.field3.append(1)


@pytest.mark.asyncio
async def test_document_update(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    await d.insert()

    await d.update(name="test2")
    result = await motor.Doc.find_one({"name": "test2"})
    assert "name" in result


@pytest.mark.asyncio
async def test_document_delete(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    await d.insert()
    await Doc.delete_all(name="test")

    docs = await Doc.get_list()
    assert len(docs) == 0


@pytest.mark.asyncio
async def test_key_escape(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    d.field4 = {"a.b": "1", "a$b": "2", "a\\b": "3", "a.b.c": 4}
    await d.insert()

    d2 = await Doc.get_by_id(d.id)

    for k in d.field4.keys():
        assert k in d2.field4


@pytest.mark.asyncio
async def test_nested_key_escape(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    d.field4["a0a7f22b-bc71-51f6-bc91-f2b0e368b786"] = {}
    d.field4["a0a7f22b-bc71-51f6-bc91-f2b0e368b786"]["changes"] = {}
    d.field4["a0a7f22b-bc71-51f6-bc91-f2b0e368b786"]["changes"]["routes"] = {}
    d.field4["a0a7f22b-bc71-51f6-bc91-f2b0e368b786"]["changes"]["routes"]["current"] = {}
    d.field4["a0a7f22b-bc71-51f6-bc91-f2b0e368b786"]["changes"]["routes"]["current"]["172.19.0.0/16"] = "1.1.1.1"
    await d.insert()

    await Doc.get_by_id(d.id)


@pytest.mark.asyncio
async def test_enum_field(motor):
    EnumDoc.set_connection(motor)

    d = EnumDoc(action=const.ResourceAction.deploy)
    await d.insert()

    new = await EnumDoc.get_by_id(d.id)

    assert new.action is const.ResourceAction.deploy
    assert new.to_dict()["action"] == const.ResourceAction.deploy


@pytest.mark.asyncio
async def test_project(data_module):
    project = data.Project(name="test")
    await project.insert()

    projects = await data.Project.get_list(name="test")
    assert len(projects) == 1
    assert projects[0].id == project.id

    other = await data.Project.get_by_id(project.id)
    assert project != other
    assert project.id == other.id


@pytest.mark.asyncio
async def test_project_unique(data_module):
    project = data.Project(name="test")
    await project.insert()

    project = data.Project(name="test")
    with pytest.raises(pymongo.errors.DuplicateKeyError):
        await project.insert()


@pytest.mark.asyncio
async def test_project_no_project_name(data_module):
    print(data_module)
    with pytest.raises(AttributeError):
        data.Project()


@pytest.mark.asyncio
async def test_project_cascade_delete(data_module):

    async def create_full_environment(project_name, environment_name):
        project = data.Project(name=project_name)
        await project.insert()

        env = data.Environment(name=environment_name, project=project.id, repo_url="", repo_branch="")
        await env.insert()

        agent_proc = data.AgentProcess(hostname="testhost",
                                       environment=env.id,
                                       first_seen=datetime.datetime.now(),
                                       last_seen=datetime.datetime.now(),
                                       sid=uuid.uuid4())
        await agent_proc.insert()

        agi1 = data.AgentInstance(process=agent_proc.id, name="agi1", tid=env.id)
        await agi1.insert()
        agi2 = data.AgentInstance(process=agent_proc.id, name="agi2", tid=env.id)
        await agi2.insert()

        agent = data.Agent(environment=env.id, name="agi1", last_failover=datetime.datetime.now(), paused=False,
                           primary=agi1.id)
        await agent.insert()

        version = int(time.time())
        cm = data.ConfigurationModel(version=version, environment=env.id)
        await cm.insert()

        resource_ids = []
        for i in range(5):
            path = "/etc/file" + str(i)
            key = "std::File[agent1,path=" + path + "]"
            res1 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                     attributes={"path": path})
            await res1.insert()
            resource_ids.append(res1.id)

        code = data.Code(version=version, resource="std::File", environment=env.id)
        await code.insert()

        unknown_parameter = data.UnknownParameter(name="test", environment=env.id, version=version, source="")
        await unknown_parameter.insert()

        return project, env, agent_proc, [agi1, agi2], agent, resource_ids, code, unknown_parameter

    async def assert_project_exists(project, env, agent_proc, agent_instances, agent, resource_ids, code, unknown_parameter,
                                    exists):
        def func(x):
            if exists:
                return x is not None
            else:
                return x is None
        assert func(await data.Project.get_by_id(project.id))
        assert func(await data.Environment.get_by_id(env.id))
        assert func(await data.AgentProcess.get_by_id(agent_proc.id))
        assert func(await data.AgentInstance.get_by_id(agent_instances[0].id))
        assert func(await data.AgentInstance.get_by_id(agent_instances[1].id))
        assert func(await data.Agent.get_by_id(agent.id))
        for current_id in resource_ids:
            assert func(await data.Resource.get_by_id(current_id))
        assert func(await data.Code.get_by_id(code.id))
        assert func(await data.UnknownParameter.get_by_id(unknown_parameter.id))

    # Setup two environments
    full_env_1 = await create_full_environment("proj1", "env1")
    full_env_2 = await create_full_environment("proj2", "env2")

    # Assert exists
    await assert_project_exists(*full_env_1, exists=True)
    await assert_project_exists(*full_env_2, exists=True)

    # Cascade delete project 1
    project_env1 = full_env_1[0]
    await project_env1.delete_cascade()

    # Assert outcome
    await assert_project_exists(*full_env_1, exists=False)
    await assert_project_exists(*full_env_2, exists=True)


@pytest.mark.asyncio
async def test_environment(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()
    assert env.project == project.id

    await project.delete_cascade()

    projects = await data.Project.get_list()
    envs = await data.Environment.get_list()
    assert len(projects) == 0
    assert len(envs) == 0


@pytest.mark.asyncio
async def test_environment_no_environment_name(data_module):
    project = data.Project(name="test")
    await project.insert()
    with pytest.raises(AttributeError):
        data.Environment(project=project.id, repo_url="", repo_branch="")


@pytest.mark.asyncio
async def test_environment_no_project_id(data_module):
    project = data.Project(name="test")
    await project.insert()
    with pytest.raises(AttributeError):
        data.Environment(name="dev", repo_url="", repo_branch="")


@pytest.mark.asyncio
async def test_environment_cascade_content_only(data_module):
    project = data.Project(name="proj")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=uuid.uuid4())
    await agent_proc.insert()

    agi1 = data.AgentInstance(process=agent_proc.id, name="agi1", tid=env.id)
    await agi1.insert()
    agi2 = data.AgentInstance(process=agent_proc.id, name="agi2", tid=env.id)
    await agi2.insert()

    agent = data.Agent(environment=env.id, name="agi1", last_failover=datetime.datetime.now(), paused=False,
                       primary=agi1.id)
    await agent.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(version=version, environment=env.id)
    await cm.insert()

    resource_ids = []
    for i in range(5):
        path = "/etc/file" + str(i)
        key = "std::File[agent1,path=" + path + "]"
        res1 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                 attributes={"path": path})
        await res1.insert()
        resource_ids.append(res1.id)

    code = data.Code(version=version, resource="std::File", environment=env.id)
    await code.insert()

    unknown_parameter = data.UnknownParameter(name="test", environment=env.id, version=version, source="")
    await unknown_parameter.insert()

    await env.set(data.AUTO_DEPLOY, True)

    assert (await data.Project.get_by_id(project.id)) is not None
    assert (await data.Environment.get_by_id(env.id)) is not None
    assert (await data.AgentProcess.get_by_id(agent_proc.id)) is not None
    assert (await data.AgentInstance.get_by_id(agi1.id)) is not None
    assert (await data.AgentInstance.get_by_id(agi2.id)) is not None
    assert (await data.Agent.get_by_id(agent.id)) is not None
    for current_id in resource_ids:
        assert (await data.Resource.get_by_id(current_id)) is not None
    assert (await data.Code.get_by_id(code.id)) is not None
    assert (await data.UnknownParameter.get_by_id(unknown_parameter.id)) is not None
    assert (await env.get(data.AUTO_DEPLOY)) is True

    await env.delete_cascade(only_content=True)

    assert (await data.Project.get_by_id(project.id)) is not None
    assert (await data.Environment.get_by_id(env.id)) is not None
    assert (await data.AgentProcess.get_by_id(agent_proc.id)) is None
    assert (await data.AgentInstance.get_by_id(agi1.id)) is None
    assert (await data.AgentInstance.get_by_id(agi2.id)) is None
    assert (await data.Agent.get_by_id(agent.id)) is None
    for current_id in resource_ids:
        assert (await data.Resource.get_by_id(current_id)) is None
    assert (await data.Code.get_by_id(code.id)) is None
    assert (await data.UnknownParameter.get_by_id(unknown_parameter.id)) is None
    assert (await env.get(data.AUTO_DEPLOY)) is True


@pytest.mark.asyncio
async def test_environment_set_setting_parameter(data_module):
    project = data.Project(name="proj")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    assert (await env.get(data.AUTO_DEPLOY)) is False
    await env.set(data.AUTO_DEPLOY, True)
    assert (await env.get(data.AUTO_DEPLOY)) is True
    await env.unset(data.AUTO_DEPLOY)
    assert (await env.get(data.AUTO_DEPLOY)) is False

    with pytest.raises(KeyError):
        await env.set("set_non_existing_parameter", 1)
    with pytest.raises(KeyError):
        await env.get("get_non_existing_parameter")
    with pytest.raises(AttributeError):
        await env.set(data.AUTO_DEPLOY, 5)


@pytest.mark.asyncio
async def test_environment_deprecated_setting(data_module, caplog):
    project = data.Project(name="proj")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    for (deprecated_option, new_option) in [(data.AUTOSTART_AGENT_INTERVAL, data.AUTOSTART_AGENT_DEPLOY_INTERVAL),
                                            (data.AUTOSTART_SPLAY, data.AUTOSTART_AGENT_DEPLOY_SPLAY_TIME)]:
        await env.set(deprecated_option, 22)
        caplog.clear()
        assert (await env.get(new_option)) == 22
        assert "Config option %s is deprecated. Use %s instead." % (deprecated_option, new_option) in caplog.text

        await env.set(new_option, 23)
        caplog.clear()
        assert (await env.get(new_option)) == 23
        assert "Config option %s is deprecated. Use %s instead." % (deprecated_option, new_option) not in caplog.text

        await env.unset(deprecated_option)
        caplog.clear()
        assert (await env.get(new_option)) == 23
        assert "Config option %s is deprecated. Use %s instead." % (deprecated_option, new_option) not in caplog.text


@pytest.mark.asyncio
async def test_agent_process(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    sid = uuid.uuid4()
    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=sid)
    await agent_proc.insert()

    agi1 = data.AgentInstance(process=agent_proc.id, name="agi1", tid=env.id)
    await agi1.insert()
    agi2 = data.AgentInstance(process=agent_proc.id, name="agi2", tid=env.id)
    await agi2.insert()

    agent_procs = await data.AgentProcess.get_by_env(env=env.id)
    assert len(agent_procs) == 1
    assert agent_procs[0].id == agent_proc.id

    assert (await data.AgentProcess.get_by_sid(sid)).id == agent_proc.id
    assert (await data.AgentProcess.get_by_sid(uuid.UUID(int=1))) is None

    live_procs = await data.AgentProcess.get_live()
    assert len(live_procs) == 1
    assert live_procs[0].id == agent_proc.id

    live_by_env_procs = await data.AgentProcess.get_live_by_env(env=env.id)
    assert len(live_by_env_procs) == 1
    assert live_by_env_procs[0].id == agent_proc.id

    await agent_proc.update_fields(expired=datetime.datetime.now())

    live_procs = await data.AgentProcess.get_live()
    assert len(live_procs) == 0

    live_by_env_procs = await data.AgentProcess.get_live_by_env(env=env.id)
    assert len(live_by_env_procs) == 0

    await agent_proc.delete_cascade()

    assert (await data.AgentProcess.get_by_id(agent_proc.id)) is None
    assert (await data.AgentInstance.get_by_id(agi1.id)) is None
    assert (await data.AgentInstance.get_by_id(agi2.id)) is None


@pytest.mark.asyncio
async def test_agent_instance(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    sid = uuid.uuid4()
    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=sid)
    await agent_proc.insert()

    agi1_name = "agi1"
    agi1 = data.AgentInstance(process=agent_proc.id, name=agi1_name, tid=env.id)
    await agi1.insert()
    agi2_name = "agi2"
    agi2 = data.AgentInstance(process=agent_proc.id, name=agi2_name, tid=env.id)
    await agi2.insert()

    active_instances = await data.AgentInstance.active()
    assert len(active_instances) == 2
    assert agi1.id in [x.id for x in active_instances]
    assert agi2.id in [x.id for x in active_instances]

    current_instances = await data.AgentInstance.active_for(env.id, agi1_name)
    assert len(current_instances) == 1
    assert current_instances[0].id == agi1.id
    current_instances = await data.AgentInstance.active_for(env.id, agi2_name)
    assert len(current_instances) == 1
    assert current_instances[0].id == agi2.id

    await agi1.update_fields(expired=datetime.datetime.now())

    active_instances = await data.AgentInstance.active()
    assert len(active_instances) == 1
    assert agi1.id not in [x.id for x in active_instances]
    assert agi2.id in [x.id for x in active_instances]

    current_instances = await data.AgentInstance.active_for(env.id, agi1_name)
    assert len(current_instances) == 0
    current_instances = await data.AgentInstance.active_for(env.id, agi2_name)
    assert len(current_instances) == 1
    assert current_instances[0].id == agi2.id


@pytest.mark.asyncio
async def test_agent(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    sid = uuid.uuid4()
    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=sid)
    await agent_proc.insert()

    agi1_name = "agi1"
    agi1 = data.AgentInstance(process=agent_proc.id, name=agi1_name, tid=env.id)
    await agi1.insert()

    agent1 = data.Agent(environment=env.id, name="agi1_agent1", last_failover=datetime.datetime.now(), paused=False,
                        primary=agi1.id)
    await agent1.insert()
    agent2 = data.Agent(environment=env.id, name="agi1_agent2", paused=False)
    await agent2.insert()
    agent3 = data.Agent(environment=env.id, name="agi1_agent3", paused=True)
    await agent3.insert()

    agents = await data.Agent.get_list()
    assert len(agents) == 3
    for agent in agents:
        assert agent.id in [agent1.id, agent2.id, agent3.id]

    for agent in [agent1, agent2, agent3]:
        retrieved_agent = await agent.get(agent.environment, agent.name)
        assert retrieved_agent is not None
        assert retrieved_agent.id == agent.id

    assert agent1.get_status() == "up"
    assert agent2.get_status() == "down"
    assert agent3.get_status() == "paused"

    for agent in [agent1, agent2, agent3]:
        assert agent.to_dict()["state"] == agent.get_status()

    await agent1.update_fields(paused=True)
    assert agent1.get_status() == "paused"

    await agent2.update_fields(primary=agi1.id)
    assert agent2.get_status() == "up"

    await agent3.update_fields(paused=False)
    assert agent3.get_status() == "down"

    primary_instance = await data.AgentInstance.get_by_id(agent1.primary)
    primary_process = await data.AgentProcess.get_by_id(primary_instance.process)
    assert primary_process.id == agent_proc.id


@pytest.mark.asyncio
async def test_config_model(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(),
                                 total=1, version_info={})
    await cm.insert()

    # create resources
    key = "std::File[agent1,path=/etc/motd]"
    res1 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version, attributes={"path": "/etc/motd"})
    await res1.insert()

    agents = await data.ConfigurationModel.get_agents(env.id, version)
    assert len(agents) == 1
    assert "agent1" in agents


@pytest.mark.asyncio
async def test_model_list(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    for version in range(1, 20):
        cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0,
                                     version_info={})
        await cm.insert()

    versions = await data.ConfigurationModel.get_versions(env.id, 0, 1)
    assert len(versions) == 1
    assert versions[0].version == 19

    versions = await data.ConfigurationModel.get_versions(env.id, 1, 1)
    assert len(versions) == 1
    assert versions[0].version == 18

    versions = await data.ConfigurationModel.get_versions(env.id)
    assert len(versions) == 19
    assert versions[0].version == 19
    assert versions[-1].version == 1

    versions = await data.ConfigurationModel.get_versions(env.id, 10)
    assert len(versions) == 9
    assert versions[0].version == 9
    assert versions[-1].version == 1


@pytest.mark.asyncio
async def test_model_get_latest_version(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    cms = []
    for version in range(1, 5):
        cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0,
                                     version_info={})
        await cm.insert()
        cms.append(cm)

    latest_version = await data.ConfigurationModel.get_latest_version(env.id)
    assert latest_version is None

    await cms[1].update_fields(released=True)
    latest_version = await data.ConfigurationModel.get_latest_version(env.id)
    assert latest_version.version == 2

    await cms[3].update_fields(released=True)
    latest_version = await data.ConfigurationModel.get_latest_version(env.id)
    assert latest_version.version == 4


@pytest.mark.asyncio
async def test_model_set_ready(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0,
                                 version_info={})
    await cm.insert()

    amount_of_done_resources = cm.done
    assert amount_of_done_resources == 0

    path = "/etc/file"
    key = "std::File[agent1,path=" + path + "]"
    resource = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                 attributes={"path": path})
    await resource.insert()

    assert cm.done == 0

    await data.ConfigurationModel.set_ready(env.id, version, resource.id, resource.resource_id,
                                            const.ResourceState.deployed)

    cm = await data.ConfigurationModel.get_by_id(cm.id)
    assert cm.done == 1


@pytest.mark.asyncio
async def test_model_delete_cascade(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0,
                                 version_info={})
    await cm.insert()

    path = "/etc/file"
    key = "std::File[agent1,path=" + path + "]"
    resource = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                 attributes={"path": path})
    await resource.insert()

    code = data.Code(version=version, resource="std::File", environment=env.id)
    await code.insert()

    unknown_parameter = data.UnknownParameter(name="test", environment=env.id, version=version, source="")
    await unknown_parameter.insert()

    await cm.delete_cascade()

    assert (await data.ConfigurationModel.get_by_id(cm.id)) is None
    assert (await data.Resource.get_by_id(resource.id)) is None
    assert (await data.Code.get_by_id(code.id)) is None
    assert (await data.UnknownParameter.get_by_id(unknown_parameter.id)) is None


@pytest.mark.asyncio
async def test_undeployable_cache_lazy(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=5, version_info={},
                                  released=False, deployed=False)
    await cm1.insert()
    await populate_model(env.id, version)

    assert cm1.undeployable is None

    undep = await cm1.get_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 3]

    assert cm1.undeployable is not None

    undep = await cm1.get_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 3]

    cm1 = await data.ConfigurationModel.get_version(env.id, version)

    assert cm1.undeployable is not None

    undep = await cm1.get_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 3]


@pytest.mark.asyncio
async def test_undeployable_skip_cache_lazy(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = 2
    cm1 = data.ConfigurationModel(environment=env.id,
                                  version=version,
                                  date=datetime.datetime.now(),
                                  total=5,
                                  version_info={},
                                  released=False, deployed=False)
    await cm1.insert()
    await populate_model(env.id, version)

    assert cm1.skipped_for_undeployable is None

    undep = await cm1.get_skipped_for_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 4, "std::File[agent1,path=/tmp/%d]" % 5]

    assert cm1.skipped_for_undeployable is not None

    undep = await cm1.get_skipped_for_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 4, "std::File[agent1,path=/tmp/%d]" % 5]

    cm1 = await data.ConfigurationModel.get_version(env.id, version)

    assert cm1.skipped_for_undeployable is not None

    undep = await cm1.get_skipped_for_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 4, "std::File[agent1,path=/tmp/%d]" % 5]


async def populate_model(env_id, version):

    def get_path(n):
        return "/tmp/%d" % n

    def get_id(n):
        return "std::File[agent1,path=/tmp/%d],v=%s" % (n, version)

    def get_resource(n, depends, status=const.ResourceState.available):
        requires = [get_id(z) for z in depends]
        return data.Resource.new(environment=env_id, resource_version_id=get_id(n),
                                 status=status,
                                 attributes={"path": get_path(n),
                                             "purge_on_delete": False,
                                             "purged": False,
                                             "requires": requires})

    res1 = get_resource(1, [])
    await res1.insert()

    res2 = get_resource(2, [1])
    await res2.insert()

    res3 = get_resource(3, [], const.ResourceState.undefined)
    await res3.insert()

    res4 = get_resource(4, [3])
    await res4.insert()

    res5 = get_resource(5, [4])
    await res5.insert()


@pytest.mark.asyncio
async def test_resource_purge_on_delete(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    # model 1
    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=2, version_info={},
                                  released=True, deployed=True)
    await cm1.insert()

    res11 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    await res11.insert()

    res12 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent2,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": True})
    await res12.insert()

    # model 2 (multiple undeployed versions)
    while version < 10:
        version += 1
        cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                      version_info={}, released=False, deployed=False)
        await cm2.insert()

        res21 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent5,path=/etc/motd],v=%s" % version,
                                  status=const.ResourceState.available,
                                  attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        await res21.insert()

    # model 3
    version += 1
    cm3 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0, version_info={})
    await cm3.insert()

    to_purge = await data.Resource.get_deleted_resources(env.id, version, set())

    assert len(to_purge) == 1
    assert to_purge[0].model == 1
    assert to_purge[0].resource_id == "std::File[agent1,path=/etc/motd]"


@pytest.mark.asyncio
async def test_issue_422(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    # model 1
    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1, version_info={},
                                  released=True, deployed=True)
    await cm1.insert()

    res11 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    await res11.insert()

    # model 2 (multiple undeployed versions)
    version += 1
    cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=False, deployed=False)
    await cm2.insert()

    res21 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.available,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    await res21.insert()

    # model 3
    version += 1
    cm3 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0, version_info={})
    await cm3.insert()

    to_purge = await data.Resource.get_deleted_resources(env.id, version, set())

    assert len(to_purge) == 1
    assert to_purge[0].model == 1
    assert to_purge[0].resource_id == "std::File[agent1,path=/etc/motd]"


@pytest.mark.asyncio
async def test_get_latest_resource(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    key = "std::File[agent1,path=/etc/motd]"
    assert (await data.Resource.get_latest_version(env.id, key)) is None

    res11 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=1", status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    await res11.insert()

    res12 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=2", status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": True})
    await res12.insert()

    res = await data.Resource.get_latest_version(env.id, key)
    assert res.model == 2


@pytest.mark.asyncio
async def test_get_resources(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    resource_ids = []
    for i in range(1, 11):
        res = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/tmp/file%d],v=1" % i,
                                status=const.ResourceState.deployed,
                                attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        await res.insert()
        resource_ids.append(res.resource_version_id)

    resources = await data.Resource.get_resources(env.id, resource_ids)
    assert len(resources) == len(resource_ids)
    assert sorted([x.resource_version_id for x in resources]) == sorted(resource_ids)

    resources = await data.Resource.get_resources(env.id, [resource_ids[0], "abcd"])
    assert len(resources) == 1

    resources = await data.Resource.get_resources(env.id, [])
    assert len(resources) == 0


@pytest.mark.asyncio
async def test_get_resources_for_version(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    resource_ids_version_one = []
    for i in range(1, 11):
        res = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/tmp/file%d],v=1" % i,
                                status=const.ResourceState.deployed,
                                attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        await res.insert()
        resource_ids_version_one.append(res.resource_version_id)

    resource_ids_version_two = []
    for i in range(11, 21):
        res = data.Resource.new(environment=env.id, resource_version_id="std::File[agent2,path=/tmp/file%d],v=2" % i,
                                status=const.ResourceState.deployed,
                                attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        await res.insert()
        resource_ids_version_two.append(res.resource_version_id)

    async def make_with_status(i, status):
        res = data.Resource.new(environment=env.id, resource_version_id="std::File[agent3,path=/tmp/file%d],v=3" % i,
                                status=status,
                                attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        await res.insert()
        return res.resource_version_id

    d = await make_with_status(1, const.ResourceState.deployed)
    s = await make_with_status(2, const.ResourceState.skipped)
    su = await make_with_status(3, const.ResourceState.skipped_for_undefined)
    u = await make_with_status(4, const.ResourceState.undefined)

    resources = await data.Resource.get_resources_for_version(env.id, 1)
    assert len(resources) == 10
    assert sorted(resource_ids_version_one) == sorted([x.resource_version_id for x in resources])
    resources = await data.Resource.get_resources_for_version(env.id, 2)
    assert len(resources) == 10
    assert sorted(resource_ids_version_two) == sorted([x.resource_version_id for x in resources])
    resources = await data.Resource.get_resources_for_version(env.id, 4)
    assert resources == []

    resources = await data.Resource.get_resources_for_version(env.id, 3)
    assert len(resources) == 4
    assert sorted([x.resource_version_id for x in resources]) == sorted([d, s, u, su])

    resources = await data.Resource.get_resources_for_version(env.id, 3, include_undefined=False)
    assert len(resources) == 2
    assert sorted([x.resource_version_id for x in resources]) == sorted([d, s])


@pytest.mark.asyncio
async def test_escaped_resources(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    routes = {"8.0.0.0/8": "1.2.3.4", "0.0.0.0/0": "127.0.0.1"}
    res = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,name=router],v=1",
                            status=const.ResourceState.deployed,
                            attributes={"name": "router", "purge_on_delete": True, "purged": False, "routes": routes})
    await res.insert()
    resource_id = res.resource_version_id

    resources = await data.Resource.get_resources(env.id, [resource_id])
    assert len(resources) == 1

    assert resources[0].attributes["routes"] == routes


@pytest.mark.asyncio
async def test_resource_provides(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    res1 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=1",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    res2 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=1",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    res1.provides.append(res2.resource_version_id)

    assert len(res1.provides) == 1
    assert len(res2.provides) == 0
    assert res1.provides[0] == res2.resource_version_id
    assert res2.provides == []

    await res1.insert()
    await res2.insert()

    res1 = await data.Resource.get(env.id, res1.resource_version_id)
    res2 = await data.Resource.get(env.id, res2.resource_version_id)

    assert len(res1.provides) == 1
    assert len(res2.provides) == 0
    assert res1.provides[0] == res2.resource_version_id
    assert res2.provides == []


@pytest.mark.asyncio
async def test_resource_hash(data_module):
    env_id = uuid.uuid4()
    res1 = data.Resource.new(environment=env_id, resource_version_id="std::File[agent1,path=/etc/file1],v=1",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    res2 = data.Resource.new(environment=env_id, resource_version_id="std::File[agent1,path=/etc/file1],v=2",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    res3 = data.Resource.new(environment=env_id, resource_version_id="std::File[agent1,path=/etc/file1],v=3",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": True})
    await res1.insert()
    await res2.insert()
    await res3.insert()

    assert res1.attribute_hash is not None
    assert res1.attribute_hash == res2.attribute_hash
    assert res3.attribute_hash is not None
    assert res1.attribute_hash != res3.attribute_hash

    readres = await data.Resource.get_resources(env_id,
                                                [res1.resource_version_id, res2.resource_version_id, res3.resource_version_id])
    res1, res2, res3 = readres

    assert res1.attribute_hash is not None
    assert res1.attribute_hash == res2.attribute_hash
    assert res3.attribute_hash is not None
    assert res1.attribute_hash != res3.attribute_hash


@pytest.mark.asyncio
async def test_resources_report(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    # model 1
    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={},
                                  released=True, deployed=True)
    await cm1.insert()

    res11 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime(2018, 7, 14, 12, 30),
                              attributes={"path": "/etc/file1"})
    await res11.insert()
    res12 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime(2018, 7, 14, 12, 30),
                              attributes={"path": "/etc/file2"})
    await res12.insert()

    # model 2
    version += 1
    cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=False, deployed=False)
    await cm2.insert()
    res21 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                              status=const.ResourceState.available,
                              attributes={"path": "/etc/file1"})
    await res21.insert()

    # model 3
    version += 1
    cm3 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=True, deployed=True)
    await cm3.insert()

    res31 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime(2018, 7, 14, 14, 30),
                              attributes={"path": "/etc/file2"})
    await res31.insert()

    report = await data.Resource.get_resources_report(env.id)
    assert len(report) == 2
    report_as_map = {x["resource_id"]: x for x in report}
    assert "std::File[agent1,path=/etc/file1]" in report_as_map
    assert "std::File[agent1,path=/etc/file2]" in report_as_map

    assert report_as_map["std::File[agent1,path=/etc/file1]"]["resource_type"] == "std::File"
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["deployed_version"] == 1
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["latest_version"] == 2
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["last_deploy"] == datetime.datetime(2018, 7, 14, 12, 30)
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["agent"] == "agent1"

    assert report_as_map["std::File[agent1,path=/etc/file1]"]["resource_type"] == "std::File"
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["deployed_version"] == 3
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["latest_version"] == 3
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["last_deploy"] == datetime.datetime(2018, 7, 14, 14, 30)
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["agent"] == "agent1"


@pytest.mark.asyncio
async def test_resource_get_requires(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    res1 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=1",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd"})
    await res1.insert()
    res2 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=1",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "requires": [res1.resource_version_id]})
    await res2.insert()

    assert (await data.Resource.get_requires(env.id, 1, res2.resource_version_id)) == []
    resources = await data.Resource.get_requires(env.id, 1, res1.resource_version_id)
    assert len(resources) == 1
    assert resources[0].id == res2.id


@pytest.mark.asyncio
async def test_resources_delete_cascade(data_module):
    project = data.Project(name="test")
    await project.insert()
    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = 1
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                 version_info={}, released=True, deployed=True)
    await cm.insert()

    res1 = data.Resource.new(environment=env.id,
                             resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                             status=const.ResourceState.deployed, last_deploy=datetime.datetime.now(),
                             attributes={"path": "/etc/file1"})
    await res1.insert()
    res2 = data.Resource.new(environment=env.id,
                             resource_version_id="std::File[agent1,path=/etc/file2],v=%s" % version,
                             status=const.ResourceState.deployed, last_deploy=datetime.datetime.now(),
                             attributes={"path": "/etc/file1"})
    await res2.insert()
    resource_action1 = data.ResourceAction(environment=env.id, resource_version_ids=[res1.resource_version_id],
                                           action_id=uuid.uuid4(), action=const.ResourceAction.deploy,
                                           started=datetime.datetime.now())
    await resource_action1.insert()
    resource_action2 = data.ResourceAction(environment=env.id, resource_version_ids=[res2.resource_version_id],
                                           action_id=uuid.uuid4(), action=const.ResourceAction.deploy,
                                           started=datetime.datetime.now())
    await resource_action2.insert()

    await res1.delete_cascade()

    assert (await data.Resource.get_by_id(res1.id)) is None
    assert (await data.ResourceAction.get_by_id(resource_action1.id)) is None
    resource = await data.Resource.get_by_id(res2.id)
    assert resource is not None
    assert resource.id == res2.id
    resource_action = await data.ResourceAction.get_by_id(resource_action2.id)
    assert resource_action is not None
    assert resource_action.id == resource_action2.id


@pytest.mark.asyncio
async def test_resource_action(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    action_id = uuid.uuid4()
    resource_action = data.ResourceAction(environment=env.id, resource_version_ids=[], action_id=action_id,
                                          action=const.ResourceAction.deploy, started=datetime.datetime.now())
    await resource_action.insert()

    resource_action.add_changes({"rid": {"field1": {"old": "a", "new": "b"}, "field2": {}}})
    await resource_action.save()

    resource_action.add_changes({"rid": {"field2": {"old": "c", "new": "d"}, "field3": {}}})
    await resource_action.save()

    resource_action.add_logs([{}, {}])
    await resource_action.save()

    resource_action.add_logs([{}, {}])
    await resource_action.save()

    ra = await data.ResourceAction.get_by_id(resource_action.id)
    assert len(ra.changes["rid"]) == 3
    assert len(ra.messages) == 4

    assert ra.changes["rid"]["field1"]["old"] == "a"
    assert ra.changes["rid"]["field1"]["new"] == "b"
    assert ra.changes["rid"]["field2"]["old"] == "c"
    assert ra.changes["rid"]["field2"]["new"] == "d"
    assert ra.changes["rid"]["field3"] == {}


@pytest.mark.asyncio
async def test_resource_action_get_logs(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    for i in range(1, 11):
        action_id = uuid.uuid4()
        resource_action = data.ResourceAction(environment=env.id, resource_version_ids=[uuid.UUID(int=1)], action_id=action_id,
                                              action=const.ResourceAction.deploy, started=datetime.datetime.now())
        await resource_action.insert()
        resource_action.add_logs(["test" + str(i)])
        await resource_action.save()

    action_id = uuid.uuid4()
    resource_action = data.ResourceAction(environment=env.id, resource_version_ids=[uuid.UUID(int=1)], action_id=action_id,
                                          action=const.ResourceAction.dryrun, started=datetime.datetime.now())
    await resource_action.insert()
    resource_action.add_logs(["log", "message"])
    await resource_action.save()

    resource_actions = await data.ResourceAction.get_log(env.id, uuid.UUID(int=1))
    assert len(resource_actions) == 11
    for i in range(len(resource_actions)):
        action = resource_actions[i]
        if i == 0:
            assert action.action == const.ResourceAction.dryrun
        else:
            assert action.action == const.ResourceAction.deploy
    resource_actions = await data.ResourceAction.get_log(env.id, uuid.UUID(int=1), const.ResourceAction.dryrun.name)
    assert len(resource_actions) == 1
    assert resource_actions[0].action == const.ResourceAction.dryrun
    assert sorted(resource_actions[0].messages) == sorted(["log", "message"])
    resource_actions = await data.ResourceAction.get_log(env.id, uuid.UUID(int=1), const.ResourceAction.deploy.name, limit=2)
    assert len(resource_actions) == 2
    assert resource_actions[0].messages == ["test10"]
    assert resource_actions[1].messages == ["test9"]


@pytest.mark.asyncio
async def test_data_document_recursion(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    now = datetime.datetime.now()
    ra = data.ResourceAction(environment=env.id, resource_version_ids=["id"], action_id=uuid.uuid4(),
                             action=const.ResourceAction.store, started=now, finished=now,
                             messages=[data.LogLine.log(logging.INFO, "Successfully stored version %(version)d",
                                                        version=2)])
    await ra.insert()


@pytest.mark.asyncio
async def test_purgelog_test(data_module):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    # ResourceAction 1
    timestamp_ra1 = datetime.datetime.now() - datetime.timedelta(days=8)
    log_line_ra1 = data.LogLine.log(logging.INFO, "Successfully stored version %(version)d", version=1)
    ra1 = data.ResourceAction(environment=env.id, resource_version_ids=["id"], action_id=uuid.uuid4(),
                              action=const.ResourceAction.store, started=timestamp_ra1, finished=datetime.datetime.now(),
                              messages=[log_line_ra1])
    await ra1.insert()

    # ResourceAction 2
    timestamp_ra2 = datetime.datetime.now() - datetime.timedelta(days=6)
    log_line_ra2 = data.LogLine.log(logging.INFO, "Successfully stored version %(version)d", version=2)
    ra2 = data.ResourceAction(environment=env.id, resource_version_ids=["id"], action_id=uuid.uuid4(),
                              action=const.ResourceAction.store, started=timestamp_ra2, finished=datetime.datetime.now(),
                              messages=[log_line_ra2])
    await ra2.insert()

    assert len(await data.ResourceAction.get_list()) == 2
    await data.ResourceAction.purge_logs()
    assert len(await data.ResourceAction.get_list()) == 1
    remaining_resource_action = (await data.ResourceAction.get_list())[0]
    assert remaining_resource_action.id == ra2.id



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import functools
import hashlib
import itertools
import logging
import warnings

import pkg_resources
from pkg_resources import DistributionNotFound
from tornado.ioloop import IOLoop
from typing import Callable, Dict, Union, Tuple, Any

LOGGER = logging.getLogger(__name__)
SALT_SIZE = 16
HASH_ROUNDS = 100000


def memoize(obj):
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        if args not in cache:
            cache[args] = obj(*args, **kwargs)
        return cache[args]

    return memoizer


def get_compiler_version() -> str:
    try:
        return pkg_resources.get_distribution("inmanta").version
    except DistributionNotFound:
        LOGGER.error(
            "Could not find version number for the inmanta compiler."
            "Is inmanta installed? Use stuptools install or setuptools dev to install."
        )
        return None


def groupby(mylist, f):
    return itertools.groupby(sorted(mylist, key=f), f)


def hash_file(content: str) -> str:
    """
        Create a hash from the given content
    """
    sha1sum = hashlib.new("sha1")
    sha1sum.update(content)

    return sha1sum.hexdigest()


def is_call_ok(result: Union[int, Tuple[int, Dict[str, Any]]]) -> bool:
    if isinstance(result, tuple):
        if len(result) == 2:
            code, reply = result
        else:
            raise Exception("Handlers for method call can only return a status code and a reply")

    else:
        code = result

    return code == 200


class Scheduler(object):
    """
        An event scheduler class
    """
    def __init__(self, name: str) -> None:
        self.name = name
        self._scheduled: Dict[Callable, object] = {}

    def add_action(self, action: Callable, interval: float, initial_delay: float = None) -> None:
        """
            Add a new action

            :param action: A function to call periodically
            :param interval: The interval between execution of actions
            :param initial_delay: Delay to the first execution, defaults to interval
        """
        if initial_delay is None:
            initial_delay = interval

        LOGGER.debug("Scheduling action %s every %d seconds with initial delay %d", action, interval, initial_delay)

        def action_function():
            LOGGER.info("Calling %s" % action)
            if action in self._scheduled:
                try:
                    action()
                except Exception:
                    LOGGER.exception("Uncaught exception while executing scheduled action")

                finally:
                    handle = IOLoop.current().call_later(interval, action_function)
                    self._scheduled[action] = handle

        handle = IOLoop.current().call_later(initial_delay, action_function)
        self._scheduled[action] = handle

    def remove(self, action: Callable) -> None:
        """
            Remove a scheduled action
        """
        if action in self._scheduled:
            IOLoop.current().remove_timeout(self._scheduled[action])
            del self._scheduled[action]

    def stop(self) -> None:
        """
            Stop the scheduler
        """
        try:
            # remove can still run during stop. That is why we loop until we get a keyerror == the dict is empty
            while True:
                action, handle = self._scheduled.popitem()
                IOLoop.current().remove_timeout(handle)
        except KeyError:
            pass

    def __del__(self) -> None:
        if len(self._scheduled) > 0:
            warnings.warn("Deleting scheduler '%s' that has not been stopped properly." % self.name)



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from typing import Dict, Sequence, List, Optional, Union  # noqa: F401
from abc import abstractmethod
import traceback
from functools import lru_cache


try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False


if TYPE_CHECKING:
    import inmanta.ast.statements  # noqa: F401
    from inmanta.ast.attribute import Attribute  # noqa: F401
    from inmanta.ast.type import Type, NamedType  # noqa: F401
    from inmanta.execute.runtime import ExecutionContext, Instance, DelayedResultVariable, ResultVariable  # noqa: F401
    from inmanta.ast.statements import Statement, AssignStatement  # noqa: F401
    from inmanta.ast.entity import Entity  # noqa: F401
    from inmanta.ast.statements.define import DefineImport, DefineEntity  # noqa: F401


class Location(object):

    def __init__(self, file: str, lnr: int) -> None:
        self.file = file
        self.lnr = lnr

    def __str__(self) -> str:
        return "%s:%d" % (self.file, self.lnr)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Location):
            return False
        return self.file == other.file and self.lnr == other.lnr

    def merge(self, other: "Location") -> "Location":
        if other is None:
            return self

        assert isinstance(other, Location)
        assert self.file == other.file

        return Location(self.file, min(self.lnr, other.lnr))


class Range(Location):

    def __init__(self, file: str, start_lnr: int, start_char: int, end_lnr: int, end_char: int) -> None:
        Location.__init__(self, file, start_lnr)
        self.start_char = start_char
        self.end_lnr = end_lnr
        self.end_char = end_char

    def merge(self, other: Location) -> Location:
        if other is None:
            return self

        assert isinstance(other, Location)
        assert self.file == other.file

        if not isinstance(other, Range):
            return Location(self.file, min(self.lnr, other.lnr))
        else:
            if other.lnr < self.lnr:
                lnr = other.lnr
                start_char = other.start_char
            elif other.lnr == self.lnr:
                lnr = other.lnr
                start_char = min(self.start_char, other.start_char)
            else:
                lnr = self.lnr
                start_char = self.start_char

            if other.end_lnr > self.end_lnr:
                end_lnr = other.end_lnr
                end_char = other.end_char
            elif other.end_lnr == self.end_lnr:
                end_lnr = other.end_lnr
                end_char = max(self.end_char, other.end_char)
            else:
                end_lnr = self.lnr
                end_char = self.end_char
            return Range(self.file, lnr, start_char, end_lnr, end_char)

    def __str__(self) -> str:
        return "%s:%d:%d" % (self.file, self.lnr, self.start_char)


class Locatable(object):

    _location: Location

    def __init__(self) -> None:
        self._location = None

    def set_location(self, location: Location) -> None:
        assert location is not None and location.lnr > 0
        self._location = location

    def get_location(self) -> Location:
        assert self._location is not None
        return self._location

    location = property(get_location, set_location)


class LocatableString(object):
    """
        A string with an attached source location.

        It is not a subtype of str, as str is not a normal class
        As such, it is very important to unwrap strings ad this object is not an actual string.

        All identifiers produced by the parser are of this type.

        The unwrapping should be done in
        1. anywhere in DefinitionStatements
        2. in the constructors of other statements
    """

    def __init__(self, value: str, location: Range, lexpos: "int", namespace: "Namespace") -> None:
        self.value = value
        self.location = location

        self.lnr = location.lnr
        self.elnr = location.end_lnr
        self.end = location.end_char
        self.start = location.start_char

        self.lexpos = lexpos
        self.namespace = namespace

    def get_value(self) -> str:
        return self.value

    def get_location(self) -> Range:
        return self.location

    def __str__(self) -> str:
        return self.value


class Anchor(object):

    def __init__(self, range: Range) -> None:
        self.range = range

    def get_range(self) -> Range:
        return self.range

    def get_location(self) -> Range:
        return self.range

    @abstractmethod
    def resolve(self) -> Location:
        raise NotImplementedError()


class TypeReferenceAnchor(Anchor):

    def __init__(self, range: Range, namespace: "Namespace", type: str) -> None:
        Anchor.__init__(self, range=range)
        self.namespace = namespace
        self.type = type

    def resolve(self) -> Location:
        t = self.namespace.get_type(self.type)
        return t.get_location()


class AttributeReferenceAnchor(Anchor):

    def __init__(self, range: Range, namespace: "Namespace", type: str, attribute: str) -> None:
        Anchor.__init__(self, range=range)
        self.namespace = namespace
        self.type = type
        self.attribute = attribute

    def resolve(self) -> Location:
        instancetype = self.namespace.get_type(self.type)
        # type check impossible atm due to import loop
        # assert isinstance(instancetype, Entity)
        return instancetype.get_attribute(self.attribute).get_location()


class Namespaced(Locatable):

    @abstractmethod
    def get_namespace(self) -> "Namespace":
        raise NotImplementedError()


class Named(Namespaced):

    @abstractmethod
    def get_full_name(self) -> str:
        raise NotImplementedError()


class Import(Locatable):

    def __init__(self, target: "Namespace") -> None:
        Locatable.__init__(self)
        self.target = target


class MockImport(Import):

    def __init__(self, target: "Namespace") -> None:
        Locatable.__init__(self)
        self.target = target


class Namespace(Namespaced):
    """
        This class models a namespace that contains defined types, modules, ...
    """

    def __init__(self, name: str, parent: "Optional[Namespace]" = None) -> None:
        Namespaced.__init__(self)
        self.__name = name
        self.__parent = parent
        self.__children = {}  # type: Dict[str,Namespace]
        self.defines_types = {}  # type: Dict[str,NamedType]
        if self.__parent is not None:
            self.visible_namespaces = {self.get_full_name(): MockImport(self)}  # type: Dict[str, Import]
            self.__parent.add_child(self)
        else:
            self.visible_namespaces = {name: MockImport(self)}  # type: Dict[str, Import]
        self.primitives = None  # type: Optional[Dict[str,Type]]
        self.scope = None  # type:  Optional[ExecutionContext]

    def set_primitives(self, primitives: "Dict[str,Type]") -> None:
        self.primitives = primitives
        for child in self.children():
            child.set_primitives(primitives)

        std = self.get_ns_from_string("std")
        assert std is not None

        self.visible_namespaces["std"] = MockImport(std)

    def get_primitives(self) -> "Dict[str,Type]":
        assert self.primitives is not None
        return self.primitives

    def get_scope(self) -> "ExecutionContext":
        assert self.scope is not None
        return self.scope

    def define_type(self, name: str, newtype: "NamedType") -> None:
        if name in self.defines_types:
            raise newtype.get_double_defined_exception(self.defines_types[name])
        self.defines_types[name] = newtype

    def import_ns(self, name: str, ns: "DefineImport") -> None:
        if name in self.visible_namespaces:
            other = self.visible_namespaces[name]
            if not isinstance(other, MockImport):
                raise DuplicateException(ns, self.visible_namespaces[name], "Two import statements have the same name")
        self.visible_namespaces[name] = ns

    def lookup(self, name: str) -> "Union[Type, ResultVariable]":
        if "::" not in name:
            return self.get_scope().direct_lookup(name)

        parts = name.rsplit("::", 1)

        if parts[0] not in self.visible_namespaces:
            raise NotFoundException(None, name, "Variable %s not found" % parts[0])

        return self.visible_namespaces[parts[0]].target.get_scope().direct_lookup(parts[1])

    def get_type(self, name: str) -> "Type":
        assert self.primitives is not None
        if "::" in name:
            parts = name.rsplit("::", 1)
            if parts[0] in self.visible_namespaces:
                ns = self.visible_namespaces[parts[0]].target
                if parts[1] in ns.defines_types:
                    return ns.defines_types[parts[1]]
                else:
                    raise TypeNotFoundException(name, ns)
            else:
                raise TypeNotFoundException(name, self)
        elif name in self.primitives:
            return self.primitives[name]
        else:
            cns = self  # type: Optional[Namespace]
            while cns is not None:
                if name in cns.defines_types:
                    return cns.defines_types[name]
                cns = cns.get_parent()
            raise TypeNotFoundException(name, self)

    def get_name(self) -> str:
        """
            Get the name of this namespace
        """
        return self.__name

    def get_full_name(self) -> str:
        """
            Get the fully qualified name of this namespace
        """
        if(self.__parent is None):
            raise Exception("Should not occur, compiler corrupt")
        if self.__parent.__parent is None:
            return self.get_name()
        return self.__parent.get_full_name() + "::" + self.get_name()

    name = property(get_name)

    def set_parent(self, parent: "Namespace") -> None:
        """
            Set the parent of this namespace. This namespace is also added to
            the child list of the parent.
        """
        self.__parent = parent
        self.__parent.add_child(self)

    def get_parent(self) -> "Optional[Namespace]":
        """
            Get the parent namespace
        """
        return self.__parent

    parent = property(get_parent, set_parent)

    def get_root(self) -> "Namespace":
        """
            Get the root
        """
        if self.__parent is None:
            return self
        return self.__parent.get_root()

    def add_child(self, child_ns: "Namespace") -> None:
        """
            Add a child to the namespace.
        """
        self.__children[child_ns.get_name()] = child_ns

    def __repr__(self) -> str:
        """
            The representation of this object
        """
        if self.__parent is not None and self.__parent.get_name() != "__root__":
            return repr(self.__parent) + "::" + self.__name

        return self.__name

    def children(self, recursive: bool=False) -> "List[Namespace]":
        """
            Get the children of this namespace
        """
        children = list(self.__children.values())
        if not recursive:
            return children

        for child in self.__children.values():
            children.extend(child.children(recursive=True))

        return children

    def get_child(self, name: str) -> "Optional[Namespace]":
        """
            Returns the child namespace with the given name or None if it does
            not exist.
        """
        if name in self.__children:
            return self.__children[name]
        return None

    def get_child_or_create(self, name: str) -> "Namespace":
        """
            Returns the child namespace with the given name or None if it does
            not exist.
        """
        if name in self.__children:
            return self.__children[name]
        out = Namespace(name, self)
        return out

    def get_ns_or_create(self, name: str) -> "Namespace":
        """
            Returns the child namespace with the given name or None if it does
            not exist.
        """
        name_parts = name.split("::")
        if len(name_parts) == 1:
            parent = self.get_root()
        else:
            preparent = self.get_root()._get_ns(name_parts[:-1])
            assert preparent is not None
            parent = preparent
        return parent.get_child_or_create(name_parts[-1])

    def get_ns_from_string(self, fqtn: str) -> "Optional[Namespace]":
        """
            Get the namespace that is referenced to in the given fully qualified
            type name.

            :param fqtn: The type name
        """
        name_parts = fqtn.split("::")
        return self.get_root()._get_ns(name_parts)

    def _get_ns(self, ns_parts: List[str]) -> "Optional[Namespace]":
        """
            Return the namespace indicated by the parts list. Each element of
            the array represents a level in the namespace hierarchy.
        """
        if len(ns_parts) == 0:
            return None
        elif len(ns_parts) == 1:
            return self.get_child(ns_parts[0])
        else:
            child = self.get_child(ns_parts[0])
            if child is None:
                return None
            return child._get_ns(ns_parts[1:])

    @lru_cache()
    def to_path(self) -> List[str]:
        """
            Return a list with the namespace path elements in it.
        """
        if self.__parent is None or self.__parent.get_name() == "__root__":
            return [self.__name]
        else:
            return self.__parent.to_path() + [self.__name]

    def get_namespace(self) -> "Namespace":
        return self

    def get_location(self) -> Location:
        return self.location


class CompilerException(Exception):
    """ Base class for exceptions generated by the compiler"""

    def __init__(self, msg: str) -> None:
        Exception.__init__(self, msg)
        self.location = None  # type: Optional[Location]
        self.msg = msg

    def set_location(self, location: Location) -> None:
        if self.location is None:
            self.location = location

    def get_message(self) -> str:
        return self.msg

    def get_location(self) -> Optional[Location]:
        return self.location

    def get_causes(self) -> "List[CompilerException]":
        return []

    def format(self) -> str:
        """Make a string representation of this particular exception """
        location = self.get_location()
        if location is not None:
            return "%s (%s)" % (self.get_message(), location)
        else:
            return self.get_message()

    def format_trace(self, indent: str="", indent_level: int=0) -> str:
        """Make a representation of this exception and its causes"""
        out = indent * indent_level + self.format()

        for cause in self.get_causes():
            part = cause.format_trace(indent=indent, indent_level=indent_level + 1)
            out += "\n" + indent * indent_level + "caused by:"
            out += "\n" + part

        return out

    def __str__(self) -> str:
        return self.format()


class RuntimeException(CompilerException):
    """Baseclass for exceptions raised by the compiler after parsing is complete."""

    def __init__(self, stmt: "Optional[Locatable]", msg: str) -> None:
        CompilerException.__init__(self, msg)
        self.stmt = None
        if stmt is not None:
            self.set_location(stmt.get_location())
            self.stmt = stmt

    def set_statement(self, stmt: "Locatable", replace: bool = True) -> None:
        for cause in self.get_causes():
            if isinstance(cause, RuntimeException):
                cause.set_statement(stmt, replace)

        if replace or self.stmt is None:
            self.set_location(stmt.get_location())
            self.stmt = stmt

    def format(self) -> str:
        """Make a string representation of this particular exception """
        if self.stmt is not None:
            return "%s (reported in %s (%s))" % (self.get_message(), self.stmt, self.get_location())
        return super(RuntimeException, self).format()


class TypeNotFoundException(RuntimeException):
    """Exception raised when a type is referenced that does not exist"""

    def __init__(self, type: str, ns: Namespace) -> None:
        RuntimeException.__init__(self, stmt=None, msg="could not find type %s in namespace %s" % (type, ns))
        self.type = type
        self.ns = ns


def stringify_exception(exn: Exception) -> str:
    if isinstance(exn, CompilerException):
        return str(exn)
    return "%s: %s" % (exn.__class__.__name__, str(exn))


class ExternalException(RuntimeException):
    """
    When a plugin call produces an exception that is not a RuntimeException,
    it is wrapped in an ExternalException to make it conform to the expected interface
    """

    def __init__(self, stmt: Locatable, msg: str, cause: Exception) -> None:
        RuntimeException.__init__(self, stmt=stmt, msg=msg)

        self.__cause__ = cause

    def get_causes(self) -> List[CompilerException]:
        return []

    def format_trace(self, indent: str="", indent_level: int=0) -> str:
        """Make a representation of this exception and its causes"""
        out = indent * indent_level + self.format()

        part = traceback.format_exception_only(self.__cause__.__class__, self.__cause__)
        out += "\n" + indent * indent_level + "caused by:\n"
        for line in part:
            out += indent * (indent_level + 1) + line

        return out


class WrappingRuntimeException(RuntimeException):
    """ Baseclass for RuntimeExceptions wrapping other RuntimeException """

    def __init__(self, stmt: Locatable, msg: str, cause: RuntimeException) -> None:
        if stmt is None and isinstance(cause, RuntimeException):
            stmt = cause.stmt

        RuntimeException.__init__(self, stmt=stmt, msg=msg)

        self.__cause__ = cause  # type: RuntimeException

    def get_causes(self) -> List[CompilerException]:
        return [self.__cause__]


class AttributeException(WrappingRuntimeException):
    """ Exception raise when an attribute could not be set, always wraps another exception """

    def __init__(self, stmt: "Locatable", instance: "Instance", attribute: str, cause: RuntimeException) -> None:
        WrappingRuntimeException.__init__(
            self, stmt=stmt, msg="Could not set attribute `%s` on instance `%s`" % (attribute, str(instance)), cause=cause)
        self.attribute = attribute
        self.instance = instance


class OptionalValueException(RuntimeException):
    """Exception raised when an optional value is accessed that has no value (and is frozen)"""

    def __init__(self, instance: "Instance", attribute: "Attribute") -> None:
        RuntimeException.__init__(self, instance, "Optional variable accessed that has no value (%s.%s)" %
                                  (instance, attribute))
        self.instance = instance
        self.attribute = attribute


class IndexException(RuntimeException):
    """Exception raised when an index definition is invalid"""
    pass


class TypingException(RuntimeException):
    """Base class for exceptions raised during the typing phase of compilation"""
    pass


class KeyException(RuntimeException):
    pass


class CycleExcpetion(TypingException):
    """Exception raised when a type is its own parent (type cycle)"""

    def __init__(self, first_type: "DefineEntity", final_name: str) -> None:
        super(CycleExcpetion, self).__init__(first_type, "")
        self.types = []  # type: List[DefineEntity]
        self.complete = False
        self.final_name = final_name

    def add(self, element: "DefineEntity") -> None:
        """Collect parent entities while traveling up the stack"""
        if(self.complete):
            return
        if element.get_full_name() == self.final_name:
            self.complete = True
        self.types.append(element)

    def get_message(self) -> str:
        trace = ",".join([x.get_full_name() for x in self.types])
        return "Entity can not be its own parent %s" % (trace)


class ModuleNotFoundException(RuntimeException):

    def __init__(self, name: str, stmt: "Statement", msg: str=None) -> None:
        if msg is None:
            msg = "could not find module %s" % name
        RuntimeException.__init__(self, stmt, msg)
        self.name = name


class NotFoundException(RuntimeException):

    def __init__(self, stmt: "Optional[Statement]", name: str, msg: "Optional[str]"=None) -> None:
        if msg is None:
            msg = "could not find value %s" % name
        RuntimeException.__init__(self, stmt, msg)
        self.name = name


class DoubleSetException(RuntimeException):

    def __init__(self,
                 stmt: "Optional[Statement]",
                 value: object,
                 location: Location,
                 newvalue: object,
                 newlocation: Location) -> None:
        self.value = value  # type: object
        self.location = location
        self.newvalue = newvalue  # type: object
        self.newlocation = newlocation
        msg = ("value set twice:\n\told value: %s\n\t\tset at %s\n\tnew value: %s\n\t\tset at %s\n"
               % (self.value, self.location, self.newvalue, self.newlocation))
        RuntimeException.__init__(self, stmt, msg)


class ModifiedAfterFreezeException(RuntimeException):

    def __init__(self,
                 rv: "DelayedResultVariable",
                 instance: "Instance",
                 attribute: "Attribute",
                 value: object,
                 location: Location,
                 reverse: bool) -> None:
        RuntimeException.__init__(self, None, "List modified after freeze")
        self.instance = instance
        self.attribute = attribute
        self.value = value
        self.location = location
        self.resultvariable = rv
        self.reverse = reverse


class DuplicateException(TypingException):
    """ Exception raise when something is defined twice """

    def __init__(self, stmt: Locatable, other: Locatable, msg: str) -> None:
        TypingException.__init__(self, stmt, msg)
        self.other = other

    def format(self) -> str:
        return "%s (original at (%s)) (duplicate at (%s))" % (self.get_message(), self.location, self.other.get_location())


class CompilerError(Exception):

    pass


class MultiException(CompilerException):
    """A single exception collecting multiple CompilerExceptions"""

    def __init__(self, others: List[CompilerException]) -> None:
        CompilerException.__init__(self, "")
        self.others = others

    def get_causes(self) -> List[CompilerException]:
        return self.others

    def format(self) -> str:
        return "Reported %d errors" % len(self.others)

    def __str__(self) -> str:
        return "Reported %d errors:\n\t" % len(self.others) + '\n\t'.join([str(e) for e in self.others])



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from inmanta import compiler


def test_anchors_basic(snippetcompiler):
    snippetcompiler.setup_for_snippet("""
entity Test:
    string a = "a"
    string b
end

entity Test2 extends Test:
    foo c
end

Test.relation [0:1] -- Test2.reverse [0:]

typedef foo as string matching /^a+$/

a = Test(b="xx")
z = a.relation
u = a.b

implementation a for Test:

end

implement Test using a
""", autostd=False)
    anchormap = compiler.anchormap()

    assert len(anchormap) == 9

    checkmap = {(r.lnr, r.start_char, r.end_char): t.lnr for r, t in anchormap}

    def verify_anchor(flnr, s, e, tolnr):
        assert checkmap[(flnr, s, e)] == tolnr

    for f, t in anchormap:
        print("%s:%d -> %s" % (f, f.end_char, t))
    verify_anchor(7, 22, 26, 2)
    verify_anchor(8, 5, 8, 13)
    verify_anchor(11, 1, 5, 2)
    verify_anchor(11, 24, 29, 7)
    verify_anchor(15, 5, 9, 2)
    verify_anchor(15, 10, 11, 4)
    verify_anchor(19, 22, 26, 2)
    verify_anchor(23, 11, 15, 2)
    verify_anchor(23, 22, 23, 19)


def test_anchors_two(snippetcompiler):
    snippetcompiler.setup_for_snippet("""
entity Test:
    list a = ["a"]
    dict b
end

a = Test(b={})
z = a.a
u = a.b

implementation a for Test:

end

implement Test using a
""", autostd=False)
    anchormap = compiler.anchormap()

    assert len(anchormap) == 5

    checkmap = {(r.lnr, r.start_char, r.end_char): t.lnr for r, t in anchormap}

    def verify_anchor(flnr, s, e, tolnr):
        assert checkmap[(flnr, s, e)] == tolnr

    for f, t in anchormap:
        print("%s:%d -> %s" % (f, f.end_char, t))
    verify_anchor(7, 5, 9, 2)
    verify_anchor(7, 10, 11, 4)
    verify_anchor(11, 22, 26, 2)
    verify_anchor(15, 22, 23, 11)
    verify_anchor(15, 11, 15, 2)
